"""
argparser/storagebox.py
"""
import textwrap


def build_storagebox_parser(argument_parser):
    """
    Builds parsing rules for storagebox-related
    command-line interface arguments.
    """
    storagebox_help = \
        "Display a list of storage box records or a single storage box record."
    storagebox_usage = "mytardis storagebox [-h] {list,get} ..."
    storagebox_parser = \
        argument_parser.model_parsers.add_parser("storagebox",
                                                 help=storagebox_help,
                                                 usage=storagebox_usage)
    storagebox_command_parsers = \
        storagebox_parser.add_subparsers(help='available commands',
                                         dest='command')

    storagebox_list_help = "Display a list of storage box records."
    storagebox_list_usage = textwrap.dedent("""\
        mytardis storagebox list
            [--limit LIMIT] [--offset OFFSET] [--order_by ORDER_BY] [--json]

          EXAMPLE

          $ mytardis storagebox list

          Model: StorageBox
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/storagebox/?format=json
          Total Count: 2
          Limit: 20
          Offset: 0

          +----+------------------------------------------------+-----------------+
          | ID |                      Name                      |   Description   |
          +====+================================================+=================+
          |  1 | local box at /opt/mytardis/develop/var/store   | Default Storage |
          +----+------------------------------------------------+-----------------+
          |  2 | local box at /opt/mytardis/develop/var/staging | Default Staging |
          +----+------------------------------------------------+-----------------+
        """)
    storagebox_command_list_parser = \
        storagebox_command_parsers.add_parser("list",
                                              help=storagebox_list_help,
                                              usage=storagebox_list_usage)
    storagebox_command_list_parser.add_argument(
        "--limit", help="Maximum number of results to return.")
    storagebox_command_list_parser.add_argument(
        "--offset",
        help="Skip this many records from the start of the result set.")
    storagebox_command_list_parser.add_argument(
        "--order_by",
        help="Order by this field.")
    storagebox_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    storagebox_get_help = "Display a single storage box record."
    storagebox_get_usage = textwrap.dedent("""\
        mytardis storagebox get [-h] [--json] storage_box_id

          EXAMPLE

          $ mytardis storagebox get 2
          +----------------------+-------------------------------------------------------------+
          |   StorageBox field   |                            Value                            |
          +======================+=============================================================+
          | ID                   | 2                                                           |
          +----------------------+-------------------------------------------------------------+
          | Name                 | local box at /opt/mytardis/develop/var/staging              |
          +----------------------+-------------------------------------------------------------+
          | Description          | Default Staging                                             |
          +----------------------+-------------------------------------------------------------+
          | Django Storage Class | tardis.tardis_portal.storage.MyTardisLocalFileSystemStorage |
          +----------------------+-------------------------------------------------------------+
          | Max Size             | 2.842e+10                                                   |
          +----------------------+-------------------------------------------------------------+
          | Status               | dirty                                                       |
          +----------------------+-------------------------------------------------------------+

          +----------------------+-----------------------------------+
          | StorageBoxOption Key |      StorageBoxOption Value       |
          +======================+===================================+
          |             location | /opt/mytardis/develop/var/staging |
          +----------------------+-----------------------------------+

          +-------------------------+--------------------------------+
          | StorageBoxAttribute Key |   StorageBoxAttribute Value    |
          +=========================+================================+
          |            scp_username | mydata                         |
          +-------------------------+--------------------------------+
          |            scp_hostname | mytardisdemo.erc.monash.edu.au |
          +-------------------------+--------------------------------+
        """)
    storagebox_command_get_parser = \
        storagebox_command_parsers.add_parser("get",
                                              help=storagebox_get_help,
                                              usage=storagebox_get_usage)
    storagebox_command_get_parser.add_argument("storage_box_id",
                                               help="The storage box ID.")
    storagebox_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")



"""
mytardisclient utils
"""
def human_readable_size_string(num):
    """
    Returns human-readable string.
    """
    num = float(num)
    for unit in ['bytes', 'KB', 'MB', 'GB']:
        if num < 1024.0 and num > -1024.0:
            return "%3.0f %s" % (num, unit)
        num /= 1024.0
    return "%3.0f %s" % (num, 'TB')



"""
argparser/config.py
"""
import textwrap


def build_config_parser(argument_parser):
    """
    'mytardis config' prompts users for settings to write to
    mytardisclient.models.config.DEFAULT_CONFIG_PATH
    """
    config_help = "Set MyTardis URL, username and API key."
    config_usage = textwrap.dedent("""\
        mytardis config [-h] [--get KEY]

          EXAMPLE

          $ mytardis config
          MyTardis URL? http://mytardisdemo.erc.monash.edu.au
          MyTardis Username? demofacility
          MyTardis API key? 644be179cc6773c30fc471bad61b50c90897146c

          Wrote settings to /Users/wettenhj/.config/mytardisclient/mytardisclient.cfg

          $ mytardis config --get url
          http://mytardisdemo.erc.monash.edu.au

          $ mytardis config --get username
          demofacility

          $ mytardis config --get apikey
          644be179cc6773c30fc471bad61b50c90897146c

          $ mytardis config --get path
          /Users/wettenhj/.config/mytardisclient/mytardisclient.cfg
        """)
    config_command_parser = \
        argument_parser.model_parsers.add_parser("config", help=config_help,
                                                 usage=config_usage)
    config_command_parser.add_argument(
        "--get", dest='key',
        help="Retrieve configuration value from "
        "~/.config/mytardisclient/mytardisclient.cfg")



"""
Model class for MyTardis API v1's endpoints.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py

This following lists all of the supported endpoints:
    /api/v1/?format=json

And this lists supported methods on an endpoint:
    /api/v1/facility/schema/?format=json

The 'schema' request above requires authentication.
"""

import requests

from mytardisclient.conf import config


class ApiEndpoint(object):
    """
    Model class for MyTardis API v1's endpoints.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, model, endpoint_json):
        self.json = endpoint_json
        self.model = model
        self.list_endpoint = endpoint_json['list_endpoint']
        self.schema = endpoint_json['schema']

    def __unicode__(self):
        return "%s: %s, %s" % (self.model, self.list_endpoint, self.schema)

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return self.__unicode__()

    @staticmethod
    @config.region.cache_on_arguments(namespace="ApiEndpoint")
    def list():
        """
        Get a list of API endpoints.
        """
        url = "%s/api/v1/?format=json" % config.url
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        endpoints_json = response.json()
        return ApiEndpoints(endpoints_json)

class ApiSchema(object):
    """
    Model class for MyTardis API v1's schemas.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods
    def __init__(self, model, schema_json):
        self.model = model
        self.json = schema_json
        self.fields = schema_json['fields']
        self.filtering = schema_json['filtering'] if 'filtering' in schema_json else {}
        for key, val in self.filtering.iteritems():
            if val == 1:
                self.filtering[key] = "ALL"
            elif val == 2:
                self.filtering[key] = "ALL_WITH_RELATIONS"
        self.ordering = schema_json['ordering'] if 'ordering' in schema_json else {}
        self.allowed_list_http_methods = schema_json['allowed_list_http_methods']
        self.allowed_detail_http_methods = schema_json['allowed_detail_http_methods']
        self.default_format = schema_json['default_format']
        self.default_limit = schema_json['default_limit']

    @staticmethod
    @config.region.cache_on_arguments(namespace="ApiEndpoint")
    def get(model):
        """
        Get a list of API-accessible functionality for a particular model.
        """
        if model == "datafile":
            model = "dataset_file"
        url = "%s/api/v1/%s/schema/?format=json" % (config.url, model)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "HTTP %d" % response.status_code
            print "url: " + url
            message = response.text
            raise Exception(message)

        api_schema = response.json()
        return ApiSchema(model, api_schema)


class ApiEndpoints(object):
    """
    Dictionary of API endpoints (list_endpoint and schema)
    with model names as keys.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, json):
        """
        Dictionary of API endpoints with model names as keys.
        """
        self.json = json
        self.index = -1
        self.total_count = len(self.json.keys())

    def __len__(self):
        """
        Return number of models accessible via API.
        """
        return len(self.json.keys())

    def __getitem__(self, model):
        """
        Get an endpoint from the set.
        """
        return ApiEndpoint(model, self.json[model])

    def __iter__(self):
        """__iter__"""
        return self

    def next(self):
        """next"""
        self.index += 1
        if self.index >= len(self):
            raise StopIteration
        model = self.json.keys()[self.index]
        return ApiEndpoint(model, self.json[model])



"""
Model class for MyTardis API v1's FacilityResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests

from mytardisclient.conf import config
from .resultset import ResultSet
from .group import Group
from mytardisclient.utils.exceptions import DoesNotExist


class Facility(object):
    """
    Model class for MyTardis API v1's FacilityResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    def __init__(self, facility_json):
        self.id = facility_json['id']  # pylint: disable=invalid-name
        self.name = facility_json['name']
        self.json = facility_json
        self.manager_group = \
            Group(group_json=facility_json['manager_group'])

    def __str__(self):
        return self.name

    @staticmethod
    @config.region.cache_on_arguments(namespace="Facility")
    def list(limit=None, offset=None, order_by=None):
        """
        Get facilities I have access to
        """
        url = config.url + "/api/v1/facility/?format=json"
        if limit:
            url += "&limit=%s" % limit
        if offset:
            url += "&offset=%s" % offset
        if order_by:
            url += "&order_by=%s" % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        if limit or offset:
            filters = dict(limit=limit, offset=offset)
            return ResultSet(Facility, url, response.json(),
                             **filters)
        else:
            return ResultSet(Facility, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="Facility")
    def get(facility_id):
        """
        Get facility with id facility_id
        """
        url = "%s/api/v1/facility/?format=json&id=%s" % (config.url,
                                                         facility_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        facilities_json = response.json()
        if facilities_json['meta']['total_count'] == 0:
            message = "Facility matching filter doesn't exist."
            raise DoesNotExist(message, url, response, Facility)
        return Facility(facility_json=facilities_json['objects'][0])



#!/usr/bin/env python
"""
client.py
"""
import os
import sys

from mytardisclient import __version__ as VERSION
from mytardisclient.models.config import Config
from mytardisclient.models.config import DEFAULT_CONFIG_PATH
from mytardisclient.controllers.api import ApiController
from mytardisclient.controllers.config import ConfigController
from mytardisclient.controllers.facility import FacilityController
from mytardisclient.controllers.instrument import InstrumentController
from mytardisclient.controllers.experiment import ExperimentController
from mytardisclient.controllers.dataset import DatasetController
from mytardisclient.controllers.datafile import DataFileController
from mytardisclient.controllers.storagebox import StorageBoxController
from mytardisclient.controllers.schema import SchemaController
from mytardisclient.argparser import ArgParser


def run():
    """
    Main function for command-line interface.
    """
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    # pylint: disable=too-many-branches

    args = ArgParser().get_args()

    if args.model == 'version':
        print "MyTardis Client v%s" % VERSION
        sys.exit(0)
    if args.verbose and (not hasattr(args, 'json') or not args.json):
        print "MyTardis Client v%s" % VERSION

    config_path = DEFAULT_CONFIG_PATH
    if not os.path.exists(config_path) or \
            args.model == 'config':
        ConfigController(config_path).configure(args)
        if args.model == 'config':
            sys.exit(0)
    config = Config(config_path)
    config.validate()

    if args.verbose and (not hasattr(args, 'json') or not args.json):
        print "Config: %s" % config_path
        print "MyTardis URL: %s" % config.url
        print "Username: %s" % config.username

    if args.model == 'api':
        ApiController().run_command(args)
    elif args.model == 'facility':
        FacilityController().run_command(args)
    elif args.model == 'instrument':
        InstrumentController().run_command(args)
    elif args.model == 'experiment':
        ExperimentController().run_command(args)
    elif args.model == 'dataset':
        DatasetController().run_command(args)
    elif args.model == 'datafile':
        DataFileController().run_command(args)
    elif args.model == 'storagebox':
        StorageBoxController().run_command(args)
    elif args.model == 'schema':
        SchemaController().run_command(args)


if __name__ == "__main__":
    run()



"""
argparser/api.py
"""
import textwrap


def build_api_parser(argument_parser):
    """
    'mytardis api' allows the user to list API endpoints
    supported by the MyTardis API.
    """
    api_help = "List models accessible via MyTardis's REST API."
    api_parser = argument_parser.model_parsers.add_parser("api", help=api_help)
    api_command_parsers = \
        api_parser.add_subparsers(help='available commands',
                                  dest='command')

    api_list_help = "List models accessible via MyTardis's REST API."
    api_list_usage = textwrap.dedent("""\
        mytardis api list [-h] [--json]

          EXAMPLE

          $ mytardis api list

          API Endpoints
          +------------+---------------------+----------------------------+
          | Model      | List Endpoint       | Schema                     |
          +============+=====================+============================+
          | facility   | /api/v1/facility/   | /api/v1/facility/schema/   |
          +------------+---------------------+----------------------------+
          | instrument | /api/v1/instrument/ | /api/v1/instrument/schema/ |
          +------------+---------------------+----------------------------+
          | experiment | /api/v1/experiment/ | /api/v1/experiment/schema/ |
          +------------+---------------------+----------------------------+
          | dataset    | /api/v1/dataset/    | /api/v1/dataset/schema/    |
          +------------+---------------------+----------------------------+
           ...          ...                   ...
          +------------+---------------------+----------------------------+
        """)
    api_command_list_parser = \
        api_command_parsers.add_parser("list", help=api_list_help,
                                       usage=api_list_usage)
    api_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    api_get_help = (
        "Display the schema for a particular MyTardis API model, "
        "showing which fields are accesible via the API, which fields "
        "support filtering, and which fields support ordering.")
    api_get_usage = textwrap.dedent("""\
        mytardis api get [-h] [--json] api_model

          EXAMPLE

          $ mytardis api get facility
          +------------------+------------------------------------------+
          | API Schema field |                   Value                  |
          +==================+==========================================+
          | Model            | facility                                 |
          +------------------+------------------------------------------+
          | Fields           | id                                       |
          |                  | manager_group                            |
          |                  | name                                     |
          |                  | resource_uri                             |
          +------------------+------------------------------------------+
          | Filtering        | {                                        |
          |                  |   "id": [                                |
          |                  |     "exact"                              |
          |                  |   ],                                     |
          |                  |   "manager_group": "ALL_WITH_RELATIONS", |
          |                  |   "name": [                              |
          |                  |     "exact"                              |
          |                  |   ]                                      |
          |                  | }                                        |
          +------------------+------------------------------------------+
          | Ordering         | {}                                       |
          +------------------+------------------------------------------+
        """)
    api_command_get_parser = \
        api_command_parsers.add_parser("get", help=api_get_help,
                                       usage=api_get_usage)
    api_command_get_parser.add_argument("api_model", help="The model name.")
    api_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")



"""
Model class for MyTardis API v1's GroupResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

# pylint: disable=missing-docstring

import requests
import urllib

from mytardisclient.conf import config
from mytardisclient.logs import logger
from mytardisclient.utils.exceptions import DoesNotExist


class Group(object):
    """
    Model class for MyTardis API v1's GroupResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, name=None, group_json=None):
        self.group_id = None
        self.name = name
        self.group_json = group_json

        if group_json is not None:
            self.group_id = group_json['id']
            if name is None:
                self.name = group_json['name']

    def __str__(self):
        return self.name

    @staticmethod
    def get_group_by_name(name):
        url = config.url + "/api/v1/group/?format=json&name=" + \
            urllib.quote(name)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            logger.debug("Failed to look up group record for name \"" +
                         name + "\".")
            logger.debug(response.text)
            raise Exception(response.text)
        groups_json = response.json()
        num_groups_found = groups_json['meta']['total_count']

        if num_groups_found == 0:
            raise DoesNotExist(
                message="Group \"%s\" was not found in MyTardis" % name,
                url=url, response=response)
        else:
            logger.debug("Found group record for name '" + name + "'.")
            return Group(name=name, group_json=groups_json['objects'][0])



"""
Model class for MyTardis API v1's ExperimentResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests
import json
import os

from .resultset import ResultSet
from .schema import Schema
from .schema import ParameterName
from mytardisclient.conf import config
from mytardisclient.utils.exceptions import DoesNotExist


class Experiment(object):
    """
    Model class for MyTardis API v1's ExperimentResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    def __init__(self, experiment_json=None):
        self.json = experiment_json
        self.id = None  # pylint: disable=invalid-name
        self.title = None
        self.description = None
        self.institution_name = None
        if experiment_json:
            for key in self.__dict__.keys():
                if key in experiment_json:
                    self.__dict__[key] = experiment_json[key]
        self.parameter_sets = []
        for exp_param_set_json in experiment_json['parameter_sets']:
            self.parameter_sets.append(
                ExperimentParameterSet(exp_param_set_json))

    @staticmethod
    @config.region.cache_on_arguments(namespace="Experiment")
    def list(limit=None, offset=None, order_by=None):
        """
        Get experiments I have access to
        """
        url = config.url + "/api/v1/experiment/?format=json"
        if limit:
            url += "&limit=%s" % limit
        if offset:
            url += "&offset=%s" % offset
        if order_by:
            url += "&order_by=%s" % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)

        if limit or offset:
            filters = dict(limit=limit, offset=offset)
            return ResultSet(Experiment, url, response.json(), **filters)
        else:
            return ResultSet(Experiment, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="Experiment")
    def get(exp_id):
        """
        Get experiment with id exp_id
        """
        url = "%s/api/v1/experiment/?format=json&id=%s" \
            % (config.url, exp_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        experiments_json = response.json()
        if experiments_json['meta']['total_count'] == 0:
            message = "Experiment matching filter doesn't exist."
            raise DoesNotExist(message, url, response, Experiment)
        return Experiment(experiment_json=experiments_json['objects'][0])

    @staticmethod
    def create(title, description="", institution=None, params_file_json=None):
        """
        Create an experiment.
        """
        new_exp_json = {
            "title": title,
            "description": description,
            "immutable": False
        }
        if institution:
            new_exp_json['institution'] = institution
        if params_file_json:
            assert os.path.exists(params_file_json)
            with open(params_file_json) as params_file:
                new_exp_json['parameter_sets'] = json.load(params_file)
        url = config.url + "/api/v1/experiment/"
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_exp_json))
        if response.status_code != 201:
            message = response.text
            raise Exception(message)
        experiment_json = response.json()
        return Experiment(experiment_json)

    @staticmethod
    def update(experiment_id, title, description):
        """
        Update an experiment record.
        """
        updated_fields_json = dict()
        updated_fields_json['title'] = title
        updated_fields_json['description'] = description
        url = "%s/api/v1/experiment/%s/" % \
            (config.url, experiment_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        if response.status_code != 202:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)
        experiment_json = response.json()
        return Experiment(experiment_json)


class ExperimentParameterSet(object):
    """
    Model class for MyTardis API v1's ExperimentParameterSetResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, expparamset_json):
        self.json = expparamset_json
        self.id = expparamset_json['id']  # pylint: disable=invalid-name
        self.experiment = expparamset_json['experiment']
        self.schema = Schema(expparamset_json['schema'])
        self.parameters = []
        for exp_param_json in expparamset_json['parameters']:
            self.parameters.append(ExperimentParameter(exp_param_json))

    @staticmethod
    @config.region.cache_on_arguments(namespace="ExperimentParameterSet")
    def list(experiment_id):
        """
        List experiment parameter sets associated with experiment ID
        experiment_id.
        """
        url = "%s/api/v1/experimentparameterset/?format=json" % config.url
        url += "&experiments__id=%s" % experiment_id
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        filters = dict(experiment_id=experiment_id)
        return ResultSet(ExperimentParameterSet, url, response.json(),
                         **filters)


class ExperimentParameter(object):
    """
    Model class for MyTardis API v1's ExperimentParameterResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, expparam_json):
        self.json = expparam_json
        self.id = expparam_json['id']  # pylint: disable=invalid-name
        self.name = ParameterName.get(expparam_json['name'].split('/')[-2])
        self.string_value = expparam_json['string_value']
        self.numerical_value = expparam_json['numerical_value']
        self.datetime_value = expparam_json['datetime_value']
        self.link_id = expparam_json['link_id']
        self.value = expparam_json['value']

    @staticmethod
    @config.region.cache_on_arguments(namespace="ExperimentParameter")
    def list(exp_param_set):
        """
        List experiment parameter records in parameter set.
        """
        pass



"""
Views for MyTardis records.
"""

import json
from texttable import Texttable

from mytardisclient.models.api import ApiEndpoints
from mytardisclient.models.api import ApiSchema
from mytardisclient.models.facility import Facility
from mytardisclient.models.instrument import Instrument
from mytardisclient.models.experiment import Experiment
from mytardisclient.models.dataset import Dataset
from mytardisclient.models.datafile import DataFile
from mytardisclient.models.storagebox import StorageBox
from mytardisclient.models.schema import Schema
from mytardisclient.models.resultset import ResultSet
from mytardisclient.utils import human_readable_size_string


def render(data, render_format='table', display_heading=True):
    """
    Generic render function
    """
    if data.__class__ == ResultSet:
        return render_result_set(data, render_format, display_heading)
    elif data.__class__ == ApiEndpoints:
        return render_api_endpoints(data, render_format, display_heading)
    else:
        return render_single_record(data, render_format)


def render_single_record(data, render_format):
    """
    Render single record.
    """
    # pylint: disable=too-many-return-statements
    if data.__class__ == ApiSchema:
        return render_api_schema(data, render_format)
    elif data.__class__ == Facility:
        return render_facility(data, render_format)
    elif data.__class__ == Instrument:
        return render_instrument(data, render_format)
    elif data.__class__ == Experiment:
        return render_experiment(data, render_format)
    elif data.__class__ == Dataset:
        return render_dataset(data, render_format)
    elif data.__class__ == DataFile:
        return render_datafile(data, render_format)
    elif data.__class__ == StorageBox:
        return render_storage_box(data, render_format)
    elif data.__class__ == Schema:
        return render_schema(data, render_format)
    else:
        print "Class is " + data.__class__.__name__


def render_result_set(data, render_format, display_heading=True):
    """
    Render result set.
    """
    # pylint: disable=too-many-return-statements
    if data.model == Facility:
        return render_facilities(data, render_format, display_heading)
    elif data.model == Instrument:
        return render_instruments(data, render_format, display_heading)
    elif data.model == Experiment:
        return render_experiments(data, render_format, display_heading)
    elif data.model == Dataset:
        return render_datasets(data, render_format, display_heading)
    elif data.model == DataFile:
        return render_datafiles(data, render_format, display_heading)
    elif data.model == StorageBox:
        return render_storage_boxes(data, render_format, display_heading)
    elif data.model == Schema:
        return render_schemas(data, render_format, display_heading)
    else:
        print "Class is " + data.model.__name__


def render_api_schema(api_schema, render_format):
    """
    Render API schema
    """
    if render_format == 'json':
        return render_api_schema_as_json(api_schema)
    else:
        return render_api_schema_as_table(api_schema)


def render_api_schema_as_json(api_schema, indent=2, sort_keys=True):
    """
    Returns JSON representation of API schema.
    """
    return json.dumps(api_schema.json, indent=indent, sort_keys=sort_keys)


def render_api_schema_as_table(api_schema):
    """
    Returns ASCII table view of API schema.
    """
    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(["t", "t"])
    table.header(["API Schema field", "Value"])
    table.add_row(["Model", api_schema.model])
    table.add_row(["Fields",
                   "\n".join(sorted([field for field in api_schema.fields]))])
    table.add_row(["Filtering",
                   json.dumps(api_schema.filtering, indent=2, sort_keys=True)])
    table.add_row(["Ordering",
                   json.dumps(api_schema.ordering, indent=2, sort_keys=True)])
    return table.draw() + "\n"


def render_api_endpoints(api_endpoints, render_format, display_heading=True):
    """
    Render API endpoints
    """
    if render_format == 'json':
        return render_api_endpoints_as_json(api_endpoints)
    else:
        return render_api_endpoints_as_table(api_endpoints, display_heading)


def render_api_endpoints_as_json(api_endpoints, indent=2, sort_keys=True):
    """
    Returns JSON representation of api_endpoints.
    """
    return json.dumps(api_endpoints.json, indent=indent, sort_keys=sort_keys)


def render_api_endpoints_as_table(api_endpoints, display_heading=True):
    """
    Returns ASCII table view of api_endpoints..
    """
    heading = "\n" \
        "API Endpoints\n" if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["l", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["Model", "List Endpoint", "Schema"])
    for api_endpoint in api_endpoints:
        table.add_row([api_endpoint.model, api_endpoint.list_endpoint,
                       api_endpoint.schema])
    return heading + table.draw() + "\n"


def render_facility(facility, render_format, display_heading=True):
    """
    Render facility
    """
    if render_format == 'json':
        return render_facility_as_json(facility)
    else:
        return render_facility_as_table(facility, display_heading)


def render_facility_as_json(facility, indent=2, sort_keys=True):
    """
    Returns JSON representation of facility.
    """
    return json.dumps(facility.json, indent=indent, sort_keys=sort_keys)


def render_facility_as_table(facility, display_heading=True):
    """
    Returns ASCII table view of facility.
    """
    heading = "\nModel: Facility\n\n" if display_heading else ""

    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Facility field", "Value"])
    table.add_row(["ID", facility.id])
    table.add_row(["Name", facility.name])
    table.add_row(["Manager Group", facility.manager_group])
    return heading + table.draw() + "\n"


def render_facilities(facilities, render_format, display_heading=True):
    """
    Render facilities
    """
    if render_format == 'json':
        return render_facilities_as_json(facilities)
    else:
        return render_facilities_as_table(facilities, display_heading)


def render_facilities_as_json(facilities, indent=2, sort_keys=True):
    """
    Returns JSON representation of facilities.
    """
    return json.dumps(facilities.json, indent=indent, sort_keys=sort_keys)


def render_facilities_as_table(facilities, display_heading=True):
    """
    Returns ASCII table view of facilities..
    """
    heading = "\n" \
        "Model: Facility\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (facilities.url, facilities.total_count,
           facilities.limit, facilities.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Name", "Manager Group"])
    for facility in facilities:
        table.add_row([facility.id, facility.name, facility.manager_group])
    return heading + table.draw() + "\n"


def render_instrument(instrument, render_format):
    """
    Render instrument
    """
    if render_format == 'json':
        return render_instrument_as_json(instrument)
    else:
        return render_instrument_as_table(instrument)


def render_instrument_as_json(instrument, indent=2, sort_keys=True):
    """
    Returns JSON representation of instrument.
    """
    return json.dumps(instrument.json, indent=indent, sort_keys=sort_keys)


def render_instrument_as_table(instrument):
    """
    Returns ASCII table view of instrument.
    """
    instrument_table = Texttable()
    instrument_table.set_cols_align(['l', 'l'])
    instrument_table.set_cols_valign(['m', 'm'])
    instrument_table.header(["Instrument field", "Value"])
    instrument_table.add_row(["ID", instrument.id])
    instrument_table.add_row(["Name", instrument.name])
    instrument_table.add_row(["Facility", instrument.facility])
    return instrument_table.draw() + "\n"


def render_instruments(instruments, render_format, display_heading=True):
    """
    Render instruments
    """
    if render_format == 'json':
        return render_instruments_as_json(instruments)
    else:
        return render_instruments_as_table(instruments, display_heading)


def render_instruments_as_json(instruments, indent=2, sort_keys=True):
    """
    Returns JSON representation of instruments.
    """
    return json.dumps(instruments.json, indent=indent, sort_keys=sort_keys)


def render_instruments_as_table(instruments, display_heading=True):
    """
    Returns ASCII table view of instruments..
    """
    heading = "\n" \
        "Model: Instrument\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (instruments.url, instruments.total_count,
           instruments.limit, instruments.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Name", "Facility"])
    for instrument in instruments:
        table.add_row([instrument.id, instrument.name, instrument.facility])
    return heading + table.draw() + "\n"


def render_experiment(experiment, render_format):
    """
    Render experiment
    """
    if render_format == 'json':
        return render_experiment_as_json(experiment)
    else:
        return render_experiment_as_table(experiment)


def render_experiment_as_json(experiment, indent=2, sort_keys=True):
    """
    Returns JSON representation of experiment.
    """
    return json.dumps(experiment.json, indent=indent, sort_keys=sort_keys)


def render_experiment_as_table(experiment):
    """
    Returns ASCII table view of experiment.
    """
    exp_and_param_sets = ""

    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Experiment field", "Value"])
    table.add_row(["ID", experiment.id])
    table.add_row(["Institution", experiment.institution_name])
    table.add_row(["Title", experiment.title])
    table.add_row(["Description", experiment.description])
    exp_and_param_sets += table.draw() + "\n"

    for exp_param_set in experiment.parameter_sets:
        exp_and_param_sets += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
        table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
        table.header(["ExperimentParameter ID", "Schema", "Parameter Name",
                      "String Value", "Numerical Value", "Datetime Value",
                      "Link ID"])
        for exp_param in exp_param_set.parameters:
            table.add_row([exp_param.id,
                           exp_param.name.schema,
                           exp_param.name,
                           exp_param.string_value,
                           exp_param.numerical_value or '',
                           exp_param.datetime_value or '',
                           exp_param.link_id or ''])
        exp_and_param_sets += table.draw() + "\n"

    return exp_and_param_sets


def render_experiments(experiments, render_format, display_heading=True):
    """
    Render experiments
    """
    if render_format == 'json':
        return render_experiments_as_json(experiments)
    else:
        return render_experiments_as_table(experiments, display_heading)


def render_experiments_as_json(experiments, indent=2, sort_keys=True):
    """
    Returns JSON representation of experiments.
    """
    return json.dumps(experiments.json, indent=indent, sort_keys=sort_keys)


def render_experiments_as_table(experiments, display_heading=True):
    """
    Returns ASCII table view of experiments..
    """
    heading = "\n" \
        "Model: Experiment\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (experiments.url, experiments.total_count,
           experiments.limit, experiments.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Institution", "Title"])
    for experiment in experiments:
        table.add_row([experiment.id, experiment.institution_name,
                       experiment.title])
    return heading + table.draw() + "\n"


def render_dataset(dataset, render_format):
    """
    Render dataset
    """
    if render_format == 'json':
        return render_dataset_as_json(dataset)
    else:
        return render_dataset_as_table(dataset)


def render_dataset_as_json(dataset, indent=2, sort_keys=True):
    """
    Returns JSON representation of dataset.
    """
    return json.dumps(dataset.json, indent=indent, sort_keys=sort_keys)


def render_dataset_as_table(dataset):
    """
    Returns ASCII table view of dataset.
    """
    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Dataset field", "Value"])
    table.add_row(["ID", dataset.id])
    table.add_row(["Experiment(s)", "\n".join(dataset.experiments)])
    table.add_row(["Description", dataset.description])
    table.add_row(["Instrument", dataset.instrument])
    dataset_and_param_sets = table.draw() + "\n"

    for dataset_param_set in dataset.parameter_sets:
        dataset_and_param_sets += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
        table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
        table.header(["DatasetParameter ID", "Schema", "Parameter Name",
                      "String Value", "Numerical Value", "Datetime Value",
                      "Link ID"])
        for dataset_param in dataset_param_set.parameters:
            table.add_row([dataset_param.id,
                           dataset_param.name.schema,
                           dataset_param.name,
                           dataset_param.string_value,
                           dataset_param.numerical_value or '',
                           dataset_param.datetime_value or '',
                           dataset_param.link_id or ''])
        dataset_and_param_sets += table.draw() + "\n"

    return dataset_and_param_sets


def render_datasets(datasets, render_format, display_heading=True):
    """
    Render datasets
    """
    if render_format == 'json':
        return render_datasets_as_json(datasets)
    else:
        return render_datasets_as_table(datasets, display_heading)


def render_datasets_as_json(datasets, indent=2, sort_keys=True):
    """
    Returns JSON representation of datasets.
    """
    return json.dumps(datasets.json, indent=indent, sort_keys=sort_keys)


def render_datasets_as_table(datasets, display_heading=True):
    """
    Returns ASCII table view of datasets..
    """
    heading = "\n" \
        "Model: Dataset\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (datasets.url, datasets.total_count,
           datasets.limit, datasets.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm'])
    table.header(["Dataset ID", "Experiment(s)", "Description", "Instrument"])
    for dataset in datasets:
        table.add_row([dataset.id, "\n".join(dataset.experiments),
                       dataset.description, dataset.instrument])
    return heading + table.draw() + "\n"


def render_datafile(datafile, render_format):
    """
    Render datafile
    """
    if render_format == 'json':
        return render_datafile_as_json(datafile)
    else:
        return render_datafile_as_table(datafile)


def render_datafile_as_json(datafile, indent=2, sort_keys=True):
    """
    Returns JSON representation of datafile.
    """
    return json.dumps(datafile.json, indent=indent, sort_keys=sort_keys)


def render_datafile_as_table(datafile):
    """
    Returns ASCII table view of datafile.
    """
    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["DataFile field", "Value"])
    table.add_row(["ID", datafile.id])
    table.add_row(["Dataset", datafile.dataset])
    table.add_row(["Filename", datafile.filename])
    uris = [replica.uri for replica in datafile.replicas]
    table.add_row(["URI", "\n".join(uris)])
    table.add_row(["Verified", str(datafile.verified)])
    table.add_row(["Size", human_readable_size_string(datafile.size)])
    table.add_row(["MD5 Sum", datafile.md5sum])
    datafile_and_param_sets = table.draw() + "\n"

    for datafile_param_set in datafile.parameter_sets:
        datafile_and_param_sets += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
        table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
        table.header(["DataFileParameter ID", "Schema", "Parameter Name",
                      "String Value", "Numerical Value", "Datetime Value",
                      "Link ID"])
        for datafile_param in datafile_param_set.parameters:
            table.add_row([datafile_param.id,
                           datafile_param.name.schema,
                           datafile_param.name,
                           datafile_param.string_value,
                           datafile_param.numerical_value or '',
                           datafile_param.datetime_value or '',
                           datafile_param.link_id or ''])
        datafile_and_param_sets += table.draw() + "\n"

    return datafile_and_param_sets


def render_datafiles(datafiles, render_format, display_heading=True):
    """
    Render datafiles
    """
    if render_format == 'json':
        return render_datafiles_as_json(datafiles)
    else:
        return render_datafiles_as_table(datafiles, display_heading)


def render_datafiles_as_json(datafiles, indent=2, sort_keys=True):
    """
    Returns JSON representation of datafiles.
    """
    return json.dumps(datafiles.json, indent=indent, sort_keys=sort_keys)


def render_datafiles_as_table(datafiles, display_heading=True):
    """
    Returns ASCII table view of datafiles..
    """
    heading = "\n" \
        "Model: DataFile\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (datafiles.url, datafiles.total_count, datafiles.limit,
           datafiles.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
    table.header(["DataFile ID", "Directory", "Filename", "URI",
                  "Verified", "Size", "MD5 Sum"])
    for datafile in datafiles:
        uris = [replica.uri for replica in datafile.replicas]
        table.add_row([datafile.id, datafile.directory, datafile.filename,
                       "\n".join(uris), str(datafile.verified),
                       human_readable_size_string(datafile.size),
                       datafile.md5sum])
    return heading + table.draw() + "\n"


def render_storage_box(storage_box, render_format):
    """
    Render storage box
    """
    if render_format == 'json':
        return render_storage_box_as_json(storage_box)
    else:
        return render_storage_box_as_table(storage_box)


def render_storage_box_as_json(storage_box, indent=2, sort_keys=True):
    """
    Returns JSON representation of storage_box.
    """
    return json.dumps(storage_box.json, indent=indent, sort_keys=sort_keys)


def render_storage_box_as_table(storage_box):
    """
    Returns ASCII table view of storage_box.
    """
    storage_box_options_attributes = ""

    table = Texttable(max_width=0)
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["StorageBox field", "Value"])
    table.add_row(["ID", storage_box.id])
    table.add_row(["Name", storage_box.name])
    table.add_row(["Description", storage_box.description])
    table.add_row(["Django Storage Class", storage_box.django_storage_class])
    table.add_row(["Max Size", storage_box.max_size])
    table.add_row(["Status", storage_box.status])
    storage_box_options_attributes += table.draw() + "\n"

    storage_box_options_attributes += "\n"
    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["StorageBoxOption Key", "StorageBoxOption Value"])
    for option in storage_box.options:
        table.add_row([option.key, option.value])
    storage_box_options_attributes += table.draw() + "\n"

    storage_box_options_attributes += "\n"
    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["StorageBoxAttribute Key", "StorageBoxAttribute Value"])
    for attribute in storage_box.attributes:
        table.add_row([attribute.key, attribute.value])
    storage_box_options_attributes += table.draw() + "\n"

    return storage_box_options_attributes


def render_storage_boxes(storage_boxes, render_format, display_heading=True):
    """
    Render storage_boxes
    """
    if render_format == 'json':
        return render_storage_boxes_as_json(storage_boxes)
    else:
        return render_storage_boxes_as_table(storage_boxes, display_heading)


def render_storage_boxes_as_json(storage_boxes, indent=2, sort_keys=True):
    """
    Returns JSON representation of storage_boxes.
    """
    return json.dumps(storage_boxes.json, indent=indent, sort_keys=sort_keys)


def render_storage_boxes_as_table(storage_boxes, display_heading=True):
    """
    Returns ASCII table view of storage_boxes..
    """
    heading = "\n" \
        "Model: StorageBox\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (storage_boxes.url, storage_boxes.total_count,
           storage_boxes.limit,
           storage_boxes.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Name", "Description"])
    for storage_box in storage_boxes:
        table.add_row([storage_box.id, storage_box.name,
                       storage_box.description])
    return heading + table.draw() + "\n"


def render_schema(schema, render_format):
    """
    Render schema
    """
    if render_format == 'json':
        return render_schema_as_json(schema)
    else:
        return render_schema_as_table(schema)


def render_schema_as_json(schema, indent=2, sort_keys=True):
    """
    Returns JSON representation of schema.
    """
    return json.dumps(schema.json, indent=indent, sort_keys=sort_keys)


def render_schema_as_table(schema):
    """
    Returns ASCII table view of schema.
    """
    schema_parameter_names = ""

    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Schema field", "Value"])
    table.add_row(["ID", schema.id])
    table.add_row(["Name", schema.name])
    table.add_row(["Namespace", schema.namespace])
    table.add_row(["Type", schema.type])
    table.add_row(["Subtype", schema.subtype])
    table.add_row(["Immutable", str(bool(schema.immutable))])
    table.add_row(["Hidden", str(bool(schema.hidden))])
    schema_parameter_names += table.draw() + "\n"

    schema_parameter_names += "\n"
    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l', 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm', 'm', 'm', 'm'])
    table.header(["ParameterName ID", "Full Name", "Name", "Data Type",
                  "Units", "Immutable", "Is Searchable", "Order", "Choices",
                  "Comparison Type"])
    for parameter_name in schema.parameter_names:
        table.add_row([parameter_name.id,
                       parameter_name.full_name.encode('utf8', 'ignore'),
                       parameter_name.name, parameter_name.data_type,
                       parameter_name.units.encode('utf8', 'ignore'),
                       str(bool(parameter_name.immutable)),
                       str(bool(parameter_name.is_searchable)),
                       parameter_name.order,
                       parameter_name.choices,
                       parameter_name.comparison_type])
    schema_parameter_names += table.draw() + "\n"

    return schema_parameter_names


def render_schemas(schemas, render_format, display_heading=True):
    """
    Render schemas
    """
    if render_format == 'json':
        return render_schemas_as_json(schemas)
    else:
        return render_schemas_as_table(schemas, display_heading)


def render_schemas_as_json(schemas, indent=2, sort_keys=True):
    """
    Returns JSON representation of schemas.
    """
    return json.dumps(schemas.json, indent=indent, sort_keys=sort_keys)


def render_schemas_as_table(schemas, display_heading=True):
    """
    Returns ASCII table view of schemas..
    """
    heading = "\n" \
        "Model: Schema\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (schemas.url, schemas.total_count,
           schemas.limit, schemas.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
    table.header(["ID", "Name", "Namespace", "Type", "Subtype", "Immutable",
                  "Hidden"])
    for schema in schemas:
        table.add_row([schema.id, schema.name, schema.namespace,
                       schema.type, schema.subtype or '',
                       str(bool(schema.immutable)), str(bool(schema.hidden))])
    return heading + table.draw() + "\n"



"""
Custom exceptions to raise within MyTardis Client.
"""


class DuplicateKey(Exception):
    """
    Duplicate key exception.
    """
    def __init__(self, message):
        super(DuplicateKey, self).__init__(message)


class MultipleObjectsReturned(Exception):
    """
    Multiple objects returned exception.
    """
    def __init__(self, message, url=None, response=None):
        super(MultipleObjectsReturned, self).__init__(message)

        self.url = url
        self.response = response


class DoesNotExist(Exception):
    """
    Does not exist exception.
    """
    def __init__(self, message, url=None, response=None, model_class=None):
        super(DoesNotExist, self).__init__(message)

        self.url = url
        self.response = response
        self.model_class = model_class


class Unauthorized(Exception):
    """
    Unauthorized exception.
    """
    def __init__(self, message, url=None, response=None):
        super(Unauthorized, self).__init__(message)

        self.url = url
        self.response = response


class InternalServerError(Exception):
    """
    Internal server exception.
    """
    def __init__(self, message, url=None, response=None):
        super(InternalServerError, self).__init__(message)

        self.url = url
        self.response = response


class SshException(Exception):
    """
    SSH exception.
    """
    def __init__(self, message, returncode=None):
        super(SshException, self).__init__(message)
        self.returncode = returncode


class StagingHostRefusedSshConnection(SshException):
    """
    Staging host refused SSH connection exception.
    """
    def __init__(self, message):
        super(StagingHostRefusedSshConnection, self).__init__(message)


class StagingHostSshPermissionDenied(SshException):
    """
    Staging host permission denied exception.
    """
    def __init__(self, message):
        super(StagingHostSshPermissionDenied, self).__init__(message)


class SshControlMasterLimit(SshException):
    """
    Reached max number of connections (or attempted connections)
    in SSH ControlMaster (Mac OS X only).
    This usually means there is a critical problem with SSHing
    to the specified staging host.
    """
    def __init__(self, message):
        super(SshControlMasterLimit, self).__init__(message)


class ScpException(SshException):
    """
    SCP exception.
    """
    def __init__(self, message, command=None, returncode=None):
        super(ScpException, self).__init__(message)
        self.command = command
        self.returncode = returncode


# pylint: disable=interface-not-implemented
class NoActiveNetworkInterface(Exception):
    """
    No active network interface exception.
    """
    def __init__(self, message):
        super(NoActiveNetworkInterface, self).__init__(message)


class BrokenPipe(Exception):
    """
    Broken pipe exception.
    """
    def __init__(self, message):
        super(BrokenPipe, self).__init__(message)


class IncompatibleMyTardisVersion(Exception):
    """
    Incompatible MyTardis version exception.
    """
    def __init__(self, message):
        super(IncompatibleMyTardisVersion, self).__init__(message)


class PrivateKeyDoesNotExist(Exception):
    """
    Private key does not exist exception.
    """
    def __init__(self, message):
        super(PrivateKeyDoesNotExist, self).__init__(message)


class InvalidFolderStructure(Exception):
    """
    Invalid folder structure exception.
    """
    def __init__(self, message):
        super(InvalidFolderStructure, self).__init__(message)


class StorageBoxAttributeNotFound(Exception):
    """
    Storage box attribute not found exception.
    """
    def __init__(self, storageBox, key):
        message = "Key '%s' not found in attributes for storage box '%s'" \
            % (key, storageBox.GetName())
        super(StorageBoxAttributeNotFound, self).__init__(message)


class StorageBoxOptionNotFound(Exception):
    """
    Storage box option not found exception.
    """
    def __init__(self, storageBox, key):
        message = "Key '%s' not found in options for storage box '%s'" \
            % (key, storageBox.GetName())
        super(StorageBoxOptionNotFound, self).__init__(message)


class InvalidConfig(Exception):
    """
    Invalid config.
    """
    def __init__(self, message):
        super(InvalidConfig, self).__init__(message)


class MissingConfig(Exception):
    """
    Missing config.
    """
    def __init__(self, message):
        super(MissingConfig, self).__init__(message)



"""
Controller class for running commands (list, get, create, update)
on experiment records.
"""

from mytardisclient.models.dataset import Dataset
from mytardisclient.models.experiment import Experiment
from mytardisclient.views import render


class ExperimentController(object):
    """
    Controller class for running commands (list, get, create, update)
    on experiment records.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.limit, args.offset, args.order_by,
                             render_format)
        elif command == "get":
            return self.get(args.experiment_id, render_format)
        elif command == "create":
            return self.create(args.title, args.description,
                               args.institution, args.params, render_format)
        elif command == "update":
            return self.update(args.experiment_id, args.title,
                               args.description, render_format)

    def list(self, limit, offset, order_by, render_format):
        """
        Display list of experiment records.
        """
        # pylint: disable=no-self-use
        experiments = Experiment.list(limit, offset, order_by)
        print render(experiments, render_format)

    def get(self, experiment_id, render_format):
        """
        Display experiment record.
        """
        # pylint: disable=no-self-use
        experiment = Experiment.get(experiment_id)
        print render(experiment, render_format)
        if render_format == 'table':
            datasets = Dataset.list(experiment_id=experiment_id)
            print render(datasets, render_format, display_heading=False)

    def create(self, title, description, institution, params, render_format):
        """
        Create experiment record.
        """
        # pylint: disable=no-self-use
        # pylint: disable=too-many-arguments
        experiment = Experiment.create(title, description, institution, params)
        print render(experiment, render_format)
        print "Experiment created successfully."

    def update(self, experiment_id, title, description, render_format):
        """
        Update experiment record.
        """
        # pylint: disable=no-self-use
        experiment = \
            Experiment.update(experiment_id, title, description)
        print render(experiment, render_format)
        print "Experiment updated successfully."



"""
Controller class for listing API endpoints.
"""

from mytardisclient.models.api import ApiEndpoint
from mytardisclient.models.api import ApiSchema
from mytardisclient.views import render


class ApiController(object):
    """
    Controller class for listing API endpoints.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(render_format)
        elif command == "get":
            return self.get(args.api_model, render_format)

    def list(self, render_format):
        """
        Display list of API endpoints.
        """
        # pylint: disable=no-self-use
        api_endpoints = ApiEndpoint.list()
        print render(api_endpoints, render_format)

    def get(self, model, render_format):
        """
        Display schema for a model's API endpoint.
        """
        # pylint: disable=no-self-use
        api_schema = ApiSchema.get(model)
        print render(api_schema, render_format)



"""
resultset.py
"""


class ResultSet(object):
    """
    Abstraction to represent JSON returned by MyTardis API
    which includes a list of records and some meta information
    e.g. whether there are additional pages of records to
    retrieve.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, model, url, json, **filters):
        """
        Each record in the result set can be
        represented as an object of class model
        """
        self.model = model
        self.url = url
        self.json = json
        self.index = -1
        self.total_count = self.json['meta']['total_count']
        self.limit = self.json['meta']['limit']
        self.offset = self.json['meta']['offset']
        self.filters = str(filters)

    def __len__(self):
        """
        Return number of records in ResultSet
        """
        return len(self.json['objects'])

    def __getitem__(self, key):
        """
        Get a record from the query set.
        """
        return self.model(self.json['objects'][key])

    def __iter__(self):
        """__iter__"""
        return self

    def next(self):
        """next"""
        self.index += 1
        if self.index >= len(self):
            raise StopIteration
        return self.model(self.json['objects'][self.index])



"""
Controller class for setting up config file.
"""

import os
from mytardisclient.conf import config


class ConfigController(object):
    """
    Controller class for setting up config file.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, path):
        self.path = path

    def configure(self, args=None):
        """
        Configure MyTardis Client settings.
        """
        if args and hasattr(args, 'key') and args.key:
            print getattr(config, args.key)
            return

        if os.path.exists(self.path):
            print "A config file already exists at %s" % self.path
            overwrite = raw_input("Are you sure you want to overwrite it? ")
            if not overwrite.strip().lower().startswith('y'):
                return
            print ""

        config.url = raw_input("MyTardis URL? ")
        config.username = raw_input("MyTardis Username? ")
        config.apikey = raw_input("MyTardis API key? ")
        config.save(self.path)
        config.update_default_headers()
        print "\nWrote settings to %s" % self.path



"""
Custom logging for MyTardis Client allows logging to ~/.mytardisclient.log
"""

# pylint: disable=missing-docstring
# pylint: disable=bare-except


import threading
import logging
import os
import sys
import inspect
import pkgutil
import traceback


class Logger(object):
    """
    Allows logger.debug(...), logger.info(...) etc. to write to
    ~/.mytardisclient.log with custom formats.
    """
    def __init__(self, name):
        self.name = name
        self.logger_object = None
        self.logger_file_handler = None
        self.level = logging.DEBUG
        self.configure_logger()
        self.app_root_dir = \
            os.path.dirname(pkgutil.get_loader("mytardisclient.client").filename)

    def configure_logger(self):
        self.logger_object = logging.getLogger(self.name)
        self.logger_object.setLevel(self.level)

        self.log_format_string = \
            "%(asctime)s - %(moduleName)s - %(lineNumber)d - " \
            "%(functionName)s - %(currentThreadName)s - %(levelname)s - " \
            "%(message)s"

        # Send all log messages to a log file.
        self.logger_file_handler = \
            logging.FileHandler(os.path.join(os.path.expanduser("~"),
                                             ".MyData_debug_log.txt"))
        self.logger_file_handler.setLevel(self.level)
        self.logger_file_handler\
            .setFormatter(logging.Formatter(self.log_format_string))
        self.logger_object.addHandler(self.logger_file_handler)

    def set_log_filename(self, log_filename):
        self.logger_object.removeHandler(self.logger_file_handler)
        self.logger_file_handler = \
            logging.FileHandler(os.path.join(os.path.expanduser("~"),
                                             log_filename))
        self.logger_file_handler.setLevel(self.level)
        self.logger_file_handler\
            .setFormatter(logging.Formatter(self.log_format_string))
        self.logger_object.addHandler(self.logger_file_handler)

    def set_level(self, level):
        self.level = level
        self.logger_object.setLevel(self.level)
        for handler in self.logger_object.handlers:
            handler.setLevel(self.level)

    def debug(self, message):
        frame = inspect.currentframe()
        outer_frames = inspect.getouterframes(frame)[1]
        if hasattr(sys, "frozen"):
            try:
                module_name = os.path.basename(outer_frames[1])
            except:
                module_name = outer_frames[1]
        else:
            module_name = os.path.relpath(outer_frames[1], self.app_root_dir)
        extra = {'moduleName':  module_name,
                 'lineNumber': outer_frames[2],
                 'functionName': outer_frames[3],
                 'currentThreadName': threading.current_thread().name}
        self.logger_object.debug(message, extra=extra)

    def error(self, message):
        frame = inspect.currentframe()
        outer_frames = inspect.getouterframes(frame)[1]
        if hasattr(sys, "frozen"):
            try:
                module_name = os.path.basename(outer_frames[1])
            except:
                module_name = outer_frames[1]
        else:
            module_name = os.path.relpath(outer_frames[1], self.app_root_dir)
        extra = {'moduleName':  module_name,
                 'lineNumber': outer_frames[2],
                 'functionName': outer_frames[3],
                 'currentThreadName': threading.current_thread().name}
        self.logger_object.error(message, extra=extra)

    def warning(self, message):
        frame = inspect.currentframe()
        outer_frames = inspect.getouterframes(frame)[1]
        if hasattr(sys, "frozen"):
            try:
                module_name = os.path.basename(outer_frames[1])
            except:
                module_name = outer_frames[1]
        else:
            module_name = os.path.relpath(outer_frames[1], self.app_root_dir)
        extra = {'moduleName':  module_name,
                 'lineNumber': outer_frames[2],
                 'functionName': outer_frames[3],
                 'currentThreadName': threading.current_thread().name}
        self.logger_object.warning(message, extra=extra)

    def info(self, message):
        frame = inspect.currentframe()
        outer_frames = inspect.getouterframes(frame)[1]
        if hasattr(sys, "frozen"):
            try:
                module_name = os.path.basename(outer_frames[1])
            except:
                module_name = outer_frames[1]
        else:
            module_name = os.path.relpath(outer_frames[1], self.app_root_dir)
        extra = {'moduleName':  module_name,
                 'lineNumber': outer_frames[2],
                 'functionName': outer_frames[3],
                 'currentThreadName': threading.current_thread().name}
        self.logger_object.info(message, extra=extra)


logger = Logger("MyTardis Client")  # pylint: disable=invalid-name



"""
Model class for MyTardis API v1's StorageBoxResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests

from mytardisclient.conf import config
from .resultset import ResultSet


class StorageBox(object):
    """
    Model class for MyTardis API v1's StorageBoxResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, storage_box_json):
        self.id = storage_box_json['id']  # pylint: disable=invalid-name
        self.name = storage_box_json['name']
        self.description = storage_box_json['description']
        self.django_storage_class = storage_box_json['django_storage_class']
        self.max_size = storage_box_json['max_size']
        self.status = storage_box_json['status']
        self.json = storage_box_json
        self.attributes = []
        for attribute_json in storage_box_json['attributes']:
            self.attributes.append(StorageBoxAttribute(attribute_json))
        self.options = []
        for option_json in storage_box_json['options']:
            self.options.append(StorageBoxOption(option_json))

    def __str__(self):
        return self.name

    @staticmethod
    @config.region.cache_on_arguments(namespace="StorageBox")
    def list(limit=None, offset=None, order_by=None):
        """
        Get storage_boxes I have access to
        """
        url = "%s/api/v1/storagebox/?format=json" % config.url
        if limit:
            url += "&limit=%s" % limit
        if offset:
            url += "&offset=%s" % offset
        if order_by:
            url += "&order_by=%s" % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "URL: %s" % url
            print "HTTP %s" % response.status_code
            message = response.text
            raise Exception(message)

        if limit or offset:
            filters = dict(limit=limit, offset=offset)
            return ResultSet(StorageBox, url, response.json(),
                             **filters)
        else:
            return ResultSet(StorageBox, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="StorageBox")
    def get(storage_box_id):
        """
        Get storage_box with id storage_box_id
        """
        url = "%s/api/v1/storagebox/%s/?format=json" % \
            (config.url, storage_box_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        storage_box_json = response.json()
        return StorageBox(storage_box_json=storage_box_json)


class StorageBoxAttribute(object):
    """
    Model class for MyTardis API v1's StorageBoxAttributeResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, storage_box_attribute_json):
        self.key = storage_box_attribute_json['key']
        self.value = storage_box_attribute_json['value']
        self.json = storage_box_attribute_json


class StorageBoxOption(object):
    """
    Model class for MyTardis API v1's StorageBoxOptionResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, storage_box_option_json):
        self.key = storage_box_option_json['key']
        self.value = storage_box_option_json['value']
        self.json = storage_box_option_json



"""
Model class for MyTardis API v1's DataFileResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests
import mimetypes
import json
import os
import cgi
import hashlib
from datetime import datetime

from mytardisclient.conf import config
from .replica import Replica
from .dataset import Dataset
from .resultset import ResultSet
from .schema import Schema
from .schema import ParameterName
from mytardisclient.utils.exceptions import DoesNotExist


class DataFile(object):
    """
    Model class for MyTardis API v1's DataFileResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, datafile_json):
        self.json = datafile_json
        self.id = datafile_json['id']  # pylint: disable=invalid-name
        self.dataset = datafile_json['dataset']
        self.directory = datafile_json['directory']
        self.filename = datafile_json['filename']
        self.size = datafile_json['size']
        self.md5sum = datafile_json['md5sum']
        self.replicas = []
        for replica_json in datafile_json['replicas']:
            self.replicas.append(Replica(replica_json))
        self.parameter_sets = []
        for datafile_param_set_json in datafile_json['parameter_sets']:
            self.parameter_sets.append(
                DataFileParameterSet(datafile_param_set_json))

    @property
    def verified(self):
        """
        All replicas (DFOs) must be verified and there must be
        at least one replica (DFO).
        """
        if len(self.replicas) == 0:
            return False
        for replica in self.replicas:
            if not replica.verified:
                return False
        return True

    @staticmethod
    @config.region.cache_on_arguments(namespace="DataFile")
    def list(dataset_id=None, directory=None, filename=None,
             limit=None, offset=None, order_by=None):
        """
        Get datafiles I have access to
        """
        # pylint: disable=too-many-arguments
        url = "%s/api/v1/dataset_file/?format=json" % config.url
        if dataset_id:
            url += "&dataset__id=%s" % dataset_id
        if directory:
            url += "&directory=%s" % directory
        if filename:
            url += "&filename=%s" % filename
        if limit:
            url += "&limit=%s" % limit
        if offset:
            url += "&offset=%s" % offset
        if order_by:
            url += "&order_by=%s" % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)

        if dataset_id or limit or offset:
            filters = dict(dataset_id=dataset_id, limit=limit, offset=offset)
            return ResultSet(DataFile, url, response.json(),
                             **filters)
        else:
            return ResultSet(DataFile, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="DataFile")
    def get(datafile_id):
        """
        Get datafile record with id datafile_id
        """
        url = "%s/api/v1/dataset_file/%s/?format=json" % \
            (config.url, datafile_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            if response.status_code == 404:
                message = "Datafile with ID %s doesn't exist." % datafile_id
                raise DoesNotExist(message, url, response, DataFile)
            message = response.text
            raise Exception(message)

        datafile_json = response.json()
        return DataFile(datafile_json=datafile_json)

    @staticmethod
    def create(dataset_id, directory, storagebox, file_path):
        """
        Create a datafile record.

        See also: upload

        Suppose someone with username james generates a file called
        "results.dat" on a data analysis server called analyzer.example.com
        in the directory ~james/analysis/dataset1/.  User james could grant
        the MyTardis server temporary access to his account on
        analyzer.example.com and then have MyTardis copy the file(s) into
        a more permanent location.

        If james agrees to allow the MyTardis server to do so, it could
        SSHFS-mount james@analyzer.example.com:/home/james/analysis/,
        e.g. at /mnt/sshfs/james-anaylzer/

        Then user james doesn't need to upload results.dat, he just needs to
        tell MyTardis how to access it, and tell MyTardis that it is not yet
        in a permanent location.

        MyTardis's default storage box model generates datafile object
        identifiers which include a dataset description (e.g. 'dataset1')
        and a unique ID, resulting in path like 'dataset1-123/results.dat'
        for the datafile object.  Because user james doesn't want to have
        to create the 'dataset1-123' folder himself, he could entrust the
        MyTardis Client to do it for him.

        The MyTardis administrator can create a storage box for james called
        "james-analyzer" which is of type "receiving", meaning that it is a
        temporary location.  The storage box record (which only needs to be
        accessed by the MyTardis administrator) would include a StorageBoxOption
        with key 'location' and value '/mnt/sshfs/james-analyzer'.

        Once james knows the dataset ID of the dataset he wants to upload to
        (123 in this case), he can create a datafile record as follows:

        mytardis datafile create 123 --storagebox=james-analyzer ~/analysis/dataset1/results.dat

        The file_path argument (set to ~/analysis/dataset1/results.dat)
        specifies the location of 'results.dat' on the analysis server.  The
        only problem with this approach is that when verifying and/or copying
        the datafile, MyTardis will look for results.dat in
        /mnt/sshfs/james-analyzer/dataset1-123/results.dat, when in fact it is in
        /mnt/sshfs/james-analyzer/dataset1/results.dat

        A solution to this is to have the MyTardis server SSHFS-mount
        ~james/.mytardisclient/datasets/ instead, and have the MyTardis Client
        create a symbolic link in ~james/.mytardisclient/datsets/ named
        "dataset1-123" pointing to ~james/analysis/dataset1/.
        """
        if not directory:
            directory = ""
        dataset = Dataset.get(dataset_id)
        uri = os.path.join("%s-%s" % (dataset.description, dataset_id),
                           directory, os.path.basename(file_path))
        md5sum = hashlib.md5(open(file_path, 'rb').read()).hexdigest()
        replicas = [{
            "url": uri,
            "location": storagebox,
            "protocol": "file",
            "verified": False
        }]
        new_datafile_json = {
            'dataset': "/api/v1/dataset/%s/" % dataset_id,
            'filename': os.path.basename(file_path),
            'directory': directory or "",
            'md5sum': md5sum,
            'size': str(os.stat(file_path).st_size),
            'mimetype': mimetypes.guess_type(file_path)[0],
            'replicas': replicas,
            'parameter_sets': []
        }
        url = "%s/api/v1/dataset_file/" % config.url
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_datafile_json))
        if response.status_code != 201:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)
        datafile_id = response.headers['location'].split("/")[-2]
        new_datafile = DataFile.get(datafile_id)
        return new_datafile

    @staticmethod
    def download(datafile_id):
        """
        Download datafile with id datafile_id
        """
        url = "%s/api/v1/dataset_file/%s/download/" \
            % (config.url, datafile_id)
        headers = {
            "Authorization": "ApiKey %s:%s" % (config.username,
                                               config.apikey)}
        response = requests.get(url=url, headers=headers, stream=True)
        if response.status_code != 200:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)
        try:
            _, params = cgi.parse_header(
                response.headers.get('Content-Disposition', ''))
            filename = params['filename']
        except KeyError:
            print "response.headers: %s" % response.headers
            raise
        fileobj = open(filename, 'wb')
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                fileobj.write(chunk)
        print "Downloaded: %s" % filename

    @staticmethod
    def upload(dataset_id, directory, file_path):
        """
        Upload datafile to dataset with ID dataset_id,
        using HTTP POST.
        """
        url = "%s/api/v1/dataset_file/" % config.url
        created_time = datetime.fromtimestamp(
            os.stat(file_path).st_ctime).isoformat()
        md5sum = hashlib.md5(open(file_path, 'rb').read()).hexdigest()
        file_data = {"dataset": "/api/v1/dataset/%s/" % dataset_id,
                     "filename": os.path.basename(file_path),
                     "directory": "",
                     "md5sum": md5sum,
                     "size": str(os.stat(file_path).st_size),
                     "mimetype": mimetypes.guess_type(file_path)[0],
                     "created_time": created_time}
        if directory:
            file_data['directory'] = directory
        file_obj = open(file_path, 'rb')
        headers = {
            "Authorization": "ApiKey %s:%s" % (config.username,
                                               config.apikey)}
        response = requests.post(url, headers=headers,
                                 data={"json_data": json.dumps(file_data)},
                                 files={'attached_file': file_obj})
        file_obj.close()
        if response.status_code != 201:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)
        if directory:
            print "Uploaded: %s/%s" % (directory, file_path)
        else:
            print "Uploaded: %s" % file_path

    @staticmethod
    def update(datafile_id, md5sum):
        """
        Update a datafile record.

        Only the md5sum field can be updated at present.
        For a large file, its upload can commence before
        the local MD5 sum calculation is complete, i.e.
        the datafile record can be initially created with
        a bogus checksum which is later corrected.
        """
        updated_fields_json = {'md5sum': md5sum}
        url = "%s/api/v1/dataset_file/%s/" % \
            (config.url, datafile_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        if response.status_code != 202:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)
        datafile_json = response.json()
        return DataFile(datafile_json)


class DataFileParameterSet(object):
    """
    Model class for MyTardis API v1's DataFileParameterSetResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, datafile_paramset_json):
        self.json = datafile_paramset_json
        self.id = datafile_paramset_json['id']  # pylint: disable=invalid-name
        self.datafile = datafile_paramset_json['datafile']
        self.schema = Schema(datafile_paramset_json['schema'])
        self.parameters = []
        for datafile_param_json in datafile_paramset_json['parameters']:
            self.parameters.append(DataFileParameter(datafile_param_json))

    @staticmethod
    @config.region.cache_on_arguments(namespace="DataFileParameterSet")
    def list(datafile_id):
        """
        List datafile parameter sets associated with datafile ID
        datafile_id.
        """
        url = "%s/api/v1/datafileparameterset/?format=json" % config.url
        url += "&datafiles__id=%s" % datafile_id
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "HTTP %s" % response.status_code
            print "URL: %s" % url
            message = response.text
            raise Exception(message)

        filters = dict(datafile_id=datafile_id)
        return ResultSet(DataFileParameterSet, url, response.json(), **filters)


class DataFileParameter(object):
    """
    Model class for MyTardis API v1's DataFileParameterResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, datafile_param_json):
        self.json = datafile_param_json
        self.id = datafile_param_json['id']  # pylint: disable=invalid-name
        self.name = ParameterName.get(datafile_param_json['name'].split('/')[-2])
        self.string_value = datafile_param_json['string_value']
        self.numerical_value = datafile_param_json['numerical_value']
        self.datetime_value = datafile_param_json['datetime_value']
        self.link_id = datafile_param_json['link_id']
        self.value = datafile_param_json['value']

    @staticmethod
    @config.region.cache_on_arguments(namespace="DataFileParameter")
    def list(datafile_param_set):
        """
        List datafile parameter records in parameter set.
        """
        pass



"""
argparser/version.py
"""
import textwrap


def build_version_parser(argument_parser):
    """
    Displays the mytardisclient version
    """
    version_help = "Display the MyTardis Client version."
    version_usage = textwrap.dedent("""\
        mytardis version [-h]

        $ mytardis version
        MyTardis Client v0.0.1""")
    argument_parser.model_parsers.add_parser("version", help=version_help,
                                             usage=version_usage)



"""
Controller class for running commands (list, get, create, update)
on dataset records.
"""

from mytardisclient.models.dataset import Dataset
from mytardisclient.models.datafile import DataFile
from mytardisclient.views import render


class DatasetController(object):
    """
    Controller class for running commands (list, get, create, update)
    on dataset records.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.exp, args.limit,
                             args.offset, args.order_by,
                             render_format)
        elif command == "get":
            return self.get(args.dataset_id, render_format)
        elif command == "create":
            return self.create(args.experiment_id, args.description,
                               args.instrument, args.params, render_format)
        elif command == "update":
            return self.update(args.dataset_id, args.description,
                               render_format)

    def list(self, experiment_id, limit, offset, order_by, render_format):
        """
        Display list of dataset records.
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=no-self-use
        datasets = Dataset.list(experiment_id,
                                limit, offset, order_by)
        print render(datasets, render_format)

    def get(self, dataset_id, render_format):
        """
        Display dataset record.
        """
        # pylint: disable=no-self-use
        dataset = Dataset.get(dataset_id)
        print render(dataset, render_format)
        if render_format == 'table':
            datafiles = DataFile.list(dataset_id)
            print render(datafiles, render_format)

    def create(self, experiment_id, description, instrument_id, params,
               render_format):
        """
        Create dataset record.
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=no-self-use
        dataset = Dataset.create(experiment_id, description, instrument_id,
                                 params)
        print render(dataset, render_format)
        print "Dataset created successfully."

    def update(self, dataset_id, description, render_format):
        """
        Update dataset record.
        """
        # pylint: disable=no-self-use
        dataset = Dataset.update(dataset_id, description)
        print render(dataset, render_format)
        print "Dataset updated successfully."



"""
Model class for MyTardis API v1's UserResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""
import requests
import traceback
import urllib2

from mytardisclient.conf import config
from .group import Group
from mytardisclient.utils.exceptions import IncompatibleMyTardisVersion
from mytardisclient.utils.exceptions import DoesNotExist
from mytardisclient.logs import logger


class User(object):
    """
    Model class for MyTardis API v1's UserResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=missing-docstring
    # pylint: disable=too-many-instance-attributes
    def __init__(self,
                 username=None, name=None,
                 email=None, user_record_json=None):
        # pylint: disable=too-many-arguments
        self.user_id = None
        self.username = username
        self.name = name
        self.email = email
        self.groups = []
        self.user_record_json = user_record_json

        if user_record_json is not None:
            self.user_id = user_record_json['id']
            if username is None:
                self.username = user_record_json['username']
            if name is None:
                self.name = user_record_json['first_name'] + " " + \
                    user_record_json['last_name']
            if email is None:
                self.email = user_record_json['email']
            try:
                for group in user_record_json['groups']:
                    self.groups.append(Group(group_json=group))
            except KeyError:
                # 'groups' should be available in the user record's JSON
                message = "Incompatible MyTardis version" \
                    "\n\n" \
                    "You appear to be connecting to a MyTardis server whose " \
                    "MyTardis version doesn't provide some of the " \
                    "functionality required by MyData." \
                    "\n\n" \
                    "Please check that you are using the correct MyTardis " \
                    "URL and ask your MyTardis administrator to check " \
                    "that an appropriate version of MyTardis is installed."
                logger.error(traceback.format_exc())
                logger.error(message)
                raise IncompatibleMyTardisVersion(message)

    def __str__(self):
        return "User: " + self.username

    @staticmethod
    def get_user_by_username(username):
        url = config.url + "/api/v1/user/?format=json&username=" + username
        try:
            response = requests.get(url=url, headers=config.default_headers)
        except:
            raise Exception(traceback.format_exc())
        if response.status_code != 200:
            message = response.text
            raise Exception(message)
        try:
            user_records_json = response.json()
        except:
            logger.error(traceback.format_exc())
            raise
        num_user_records_found = user_records_json['meta']['total_count']

        if num_user_records_found == 0:
            raise DoesNotExist(
                message="User \"%s\" was not found in MyTardis" % username,
                url=url, response=response)
        else:
            logger.debug("Found user record for username '" + username + "'.")
            return User(username=username,
                        user_record_json=user_records_json['objects'][0])

    @staticmethod
    def get_user_by_email(email):
        url = config.url + "/api/v1/user/?format=json&email__iexact=" + \
            urllib2.quote(email)
        try:
            response = requests.get(url=url, headers=config.default_headers)
        except:
            raise Exception(traceback.format_exc())
        if response.status_code != 200:
            logger.debug(url)
            message = response.text
            raise Exception(message)
        try:
            user_records_json = response.json()
        except:
            logger.error(traceback.format_exc())
            raise
        num_user_records_found = user_records_json['meta']['total_count']

        if num_user_records_found == 0:
            raise DoesNotExist(
                message="User with email \"%s\" was not found in MyTardis"
                % email,
                url=url, response=response)
        else:
            logger.debug("Found user record for email '" + email + "'.")
            return User(user_record_json=user_records_json['objects'][0])


# pylint: disable=too-few-public-methods
class UserProfile(object):
    """
    Used with the DoesNotExist exception when a 404 from MyTardis's API
    is assumed to have been caused by a missing user profile record.
    """
    pass



"""
Model class for MyTardis API v1's ReplicaResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""


class Replica(object):
    """
    Model class for MyTardis API v1's ReplicaResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, replica_json):
        self.json = replica_json
        self.id = replica_json['id']  # pylint: disable=invalid-name
        self.uri = replica_json['uri']
        self.verified = replica_json['verified']



"""
Controller class for running commands (list, get)
on schemas.
"""

from mytardisclient.models.schema import Schema
from mytardisclient.views import render


class SchemaController(object):
    """
    Controller class for running commands (list, get)
    on schemas.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.limit, args.offset, args.order_by,
                             render_format)
        elif command == "get":
            return self.get(args.schema_id, render_format)

    def list(self, limit, offset, order_by, render_format):
        """
        Display list of schemas.
        """
        # pylint: disable=no-self-use
        schemas = Schema.list(limit, offset, order_by)
        print render(schemas, render_format)

    def get(self, schema_id, render_format):
        """
        Display schema record.
        """
        # pylint: disable=no-self-use
        schema = Schema.get(schema_id)
        print render(schema, render_format)



"""
argparser/facility.py
"""
import textwrap


def build_facility_parser(argument_parser):
    """
    Builds parsing rules for facility-related
    command-line interface arguments.
    """
    facility_help = \
        "Display a list of facility records or a single facility record."
    facility_usage = "mytardis facility [-h] {list,get} ..."
    facility_parser = \
        argument_parser.model_parsers.add_parser("facility",
                                                 help=facility_help,
                                                 usage=facility_usage)
    facility_command_parsers = \
        facility_parser.add_subparsers(help='available commands',
                                       dest='command')

    facility_list_help = "Display a list of facility records."
    facility_list_usage = textwrap.dedent("""\
        mytardis facility list
            [--limit LIMIT] [--offset OFFSET] [--order_by ORDER_BY] [--json]

          EXAMPLE

          $ mytardis facility list

          Model: Facility
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/facility/?format=json
          Total Count: 2
          Limit: 20
          Offset: 0

          +----+---------------+------------------------+
          | ID |     Name      |     Manager Group      |
          +====+===============+========================+
          |  1 | Demo Facility | demo-facility-managers |
          +----+---------------+------------------------+
          |  2 | Test Facility | test-facility-managers |
          +----+---------------+------------------------+
        """)
    facility_command_list_parser = \
        facility_command_parsers.add_parser("list", help=facility_list_help,
                                            usage=facility_list_usage)
    facility_command_list_parser.add_argument(
        "--limit", help="Maximum number of results to return.")
    facility_command_list_parser.add_argument(
        "--offset",
        help="Skip this many records from the start of the result set.")
    facility_command_list_parser.add_argument(
        "--order_by",
        help="Order by this field.")
    facility_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    facility_get_help = "Display a single facility record."
    facility_get_usage = textwrap.dedent("""\
        mytardis facility get [-h] [--json] facility_id

          EXAMPLE

          $ mytardis facility get 1

          Model: Facility

          +----------------+------------------------+
          | Facility field |         Value          |
          +================+========================+
          | ID             | 1                      |
          +----------------+------------------------+
          | Name           | Demo Facility          |
          +----------------+------------------------+
          | Manager Group  | demo-facility-managers |
          +----------------+------------------------+


          Model: Instrument
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/instrument/?format=json&facility__id=1
          Total Count: 3
          Limit: 20
          Offset: 0

          +----+-------------------------+---------------+
          | ID |          Name           |   Facility    |
          +====+=========================+===============+
          |  3 | Test Instrument         | Demo Facility |
          +----+-------------------------+---------------+
          |  4 | Beamline                | Demo Facility |
          +----+-------------------------+---------------+
          |  8 | James Test Instrument   | Demo Facility |
          +----+-------------------------+---------------+
        """)
    facility_command_get_parser = \
        facility_command_parsers.add_parser("get", help=facility_get_help,
                                            usage=facility_get_usage)
    facility_command_get_parser.add_argument("facility_id",
                                             help="The facility ID.")
    facility_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")



"""
argparser/datafile.py
"""
import textwrap


def build_datafile_parser(argument_parser):
    """
    Builds parsing rules for datafile-related command-line interface arguments.
    """
    # pylint: disable=too-many-locals
    datafile_help = \
        "Display a list of datafile records or a single datafile record."
    datafile_usage = \
        "mytardis datafile [-h] {list,get,create,update,download,upload} ..."
    datafile_parser = \
        argument_parser.model_parsers.add_parser("datafile",
                                                 help=datafile_help,
                                                 usage=datafile_usage)
    datafile_command_parsers = \
        datafile_parser.add_subparsers(help='available commands',
                                       dest='command')

    datafile_list_help = "Display a list of datafile records."
    datafile_list_usage = textwrap.dedent("""\
        mytardis datafile list
            [--dataset DATASET] [--directory DIRECTORY] [--filename FILENAME]
            [--limit LIMIT] [--offset OFFSET] [--order_by ORDER_BY] [--json]

          EXAMPLE
       
          $ mytardis datafile list --dataset 35

          Model: DataFile
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/dataset_file/?format=json&dataset__id=35
          Total Count: 2
          Limit: 20
          Offset: 0

          +-------------+-----------+-----------+-------------------------------------+----------+-----------+----------------------------------+
          | DataFile ID | Directory | Filename  |                 URI                 | Verified |   Size    |             MD5 Sum              |
          +=============+===========+===========+=====================================+==========+===========+==================================+
          |         120 |           | hello.txt | James Test Dataset 004-35/hello.txt | True     |  25 bytes | 214e52fcf7f98d9fb8588b42cfb7987f |
          +-------------+-----------+-----------+-------------------------------------+----------+-----------+----------------------------------+
          |         121 |           | test.txt  | James Test Dataset 004-35/test.txt  | False    |   8 bytes | 74119cb1c9568b96feb7fa392006e40f |
          +-------------+-----------+-----------+-------------------------------------+----------+-----------+----------------------------------+
        """)
    datafile_command_list_parser = \
        datafile_command_parsers.add_parser("list",
                                            help=datafile_list_help,
                                            usage=datafile_list_usage)
    datafile_command_list_parser.add_argument("--dataset",
                                              help="The dataset ID.")
    datafile_command_list_parser.add_argument("--directory",
                                              help="The subdirectory within the dataset.")
    datafile_command_list_parser.add_argument("--filename",
                                              help="The datafile's name.")
    datafile_command_list_parser.add_argument(
        "--limit", help="Maximum number of results to return.")
    datafile_command_list_parser.add_argument(
        "--offset",
        help="Skip this many records from the start of the result set.")
    datafile_command_list_parser.add_argument(
        "--order_by",
        help="Order by this field.")
    datafile_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    datafile_get_help = "Display a single datafile record."
    datafile_get_usage = textwrap.dedent("""\
        mytardis datafile get [--json] datafile_id

          EXAMPLE

          $ mytardis datafile get 99
          +----------------+-------------------------------------+
          | DataFile field |                Value                |
          +================+=====================================+
          | ID             | 99                                  |
          +----------------+-------------------------------------+
          | Dataset        | /api/v1/dataset/31/                 |
          +----------------+-------------------------------------+
          | Filename       | hello.txt                           |
          +----------------+-------------------------------------+
          | URI            | James Test Dataset 001-31/hello.txt |
          +----------------+-------------------------------------+
          | Verified       | True                                |
          +----------------+-------------------------------------+
          | Size           |  13 bytes                           |
          +----------------+-------------------------------------+
          | MD5 Sum        | 9af2f8218b150c351ad802c6f3d66abe    |
          +----------------+-------------------------------------+
        """)
    datafile_command_get_parser = \
        datafile_command_parsers.add_parser("get",
                                            help=datafile_get_help,
                                            usage=datafile_get_usage)
    datafile_command_get_parser.add_argument("datafile_id",
                                             help="The datafile ID.")
    datafile_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    datafile_create_help = textwrap.dedent("""\
        Create a datafile record.
        """)
    datafile_create_usage = textwrap.dedent("""\
        mytardis datafile create
            [--directory DIRECTORY] [--storagebox STORAGEBOX] dataset_id file_path

          EXAMPLE

          $ mytardis datafile create 31 path/test.txt

          Model: DataFile

          +----------------+------------------------------------+
          | DataFile field |               Value                |
          +================+====================================+
          | ID             | 119                                |
          +----------------+------------------------------------+
          | Dataset        | /api/v1/dataset/31/                |
          +----------------+------------------------------------+
          | Filename       | test.txt                           |
          +----------------+------------------------------------+
          | URI            | James Test Dataset 001-31/test.txt |
          +----------------+------------------------------------+
          | Verified       | False                              |
          +----------------+------------------------------------+
          | Size           |   5 bytes                          |
          +----------------+------------------------------------+
          | MD5 Sum        | 2205e48de5f93c784733ffcca841d2b5   |
          +----------------+------------------------------------+

          DataFile created successfully.
            """)
    datafile_command_create_parser = \
        datafile_command_parsers.add_parser(
            "create",
            help=datafile_create_help,
            usage=datafile_create_usage)
    datafile_command_create_parser.add_argument(
        "dataset_id", help="The dataset ID.")
    datafile_command_create_parser.add_argument(
        "--directory", help="The subdirectory within the dataset.")
    datafile_command_create_parser.add_argument(
        "--storagebox", help="The storage box containing the datafile.")
    datafile_command_create_parser.add_argument(
        "file_path", help="The file to be represented in the datafile record.")

    datafile_download_help = "Download a datafile."
    datafile_download_usage = textwrap.dedent("""\
        mytardis datafile download datafile_id

          EXAMPLE

          $ mytardis datafile download 99
          Downloaded: hello.txt
            """)
    datafile_cmd_download_parser = \
        datafile_command_parsers.add_parser("download",
                                            help=datafile_download_help,
                                            usage=datafile_download_usage)
    datafile_cmd_download_parser.add_argument("datafile_id",
                                              help="The datafile ID.")

    datafile_upload_help = "Upload a datafile."
    datafile_upload_usage = textwrap.dedent("""\
        mytardis datafile upload dataset_id [--directory DIRECTORY] file_path

          EXAMPLE

          We will upload datafile "new_datafile.txt" to dataset ID 35:

          $ mytardis datafile upload 35 new_datafile.txt
          Uploaded: new_datafile.txt

          Now let's display dataset 35 (including its datafiles) and
          check whether our new datafile has been uploaded:

          $ mytardis dataset get 35
          +---------------+------------------------+
          | Dataset field |         Value          |
          +===============+========================+
          | ID            | 35                     |
          +---------------+------------------------+
          | Experiment(s) | /api/v1/experiment/24/ |
          +---------------+------------------------+
          | Description   | James Test Dataset 004 |
          +---------------+------------------------+
          | Instrument    | None                   |
          +---------------+------------------------+

          +---------------------+---------------------+------------------------+-------------------------+-----------------+----------------+---------+
          | DatasetParameter ID |       Schema        |     Parameter Name     |      String Value       | Numerical Value | Datetime Value | Link ID |
          +=====================+=====================+========================+=========================+=================+================+=========+
          |                   1 | Test dataset schema | Example Parameter Name | Example Parameter Value |                 |                |         |
          +---------------------+---------------------+------------------------+-------------------------+-----------------+----------------+---------+


          Model: DataFile
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/dataset_file/?format=json&dataset__id=35
          Total Count: 3
          Limit: 20
          Offset: 0

          +-------------+-----------+------------------+--------------------------------------------+----------+-----------+----------------------------------+
          | DataFile ID | Directory |     Filename     |                    URI                     | Verified |   Size    |             MD5 Sum              |
          +=============+===========+==================+============================================+==========+===========+==================================+
          |         120 |           | hello.txt        | James Test Dataset 004-35/hello.txt        | True     |  25 bytes | 214e52fcf7f98d9fb8588b42cfb7987f |
          +-------------+-----------+------------------+--------------------------------------------+----------+-----------+----------------------------------+
          |         121 |           | test.txt         | James Test Dataset 004-35/test.txt         | False    |   8 bytes | 74119cb1c9568b96feb7fa392006e40f |
          +-------------+-----------+------------------+--------------------------------------------+----------+-----------+----------------------------------+
          |         131 |           | new_datafile.txt | James Test Dataset 004-35/new_datafile.txt | True     |  21 bytes | 2ce6aa694791def6de39b3fccead9d87 |
          +-------------+-----------+------------------+--------------------------------------------+----------+-----------+----------------------------------+
        """)
    datafile_cmd_upload_parser = \
        datafile_command_parsers.add_parser("upload",
                                            help=datafile_upload_help,
                                            usage=datafile_upload_usage)
    datafile_cmd_upload_parser.add_argument("dataset_id",
                                            help="The dataset ID.")
    datafile_cmd_upload_parser.add_argument(
        "--directory",
        help="The datafile's subdirectory within the dataset.")
    datafile_cmd_upload_parser.add_argument("file_path",
                                            help="The file to upload.")

    datafile_update_help = "Update a datafile record."
    datafile_update_usage = textwrap.dedent("""\
        mytardis datafile update [--md5sum MD5SUM] datafile_id

          EXAMPLE

          $ mytardis datafile update --md5sum "md5-sum" 99
          HTTP 401
          Traceback (most recent call last):
            ...
          Exception

          Datafile updates are not yet enabled in the MyTardis API.
          See the "update_detail" method in tardis/tardis_portal/api.py

          Being able to update the MD5 sum would be useful for really
          large datafiles where calculating the MD5 sum is time-consuming.
          Rather than waiting until the MD5 sum calculation is complete to
          begin a datafile upload, the datafile upload could begin
          immediately (supplying a bogus MD5 sum to MyTardis), and then
          the MD5 sum could be updated once the calculation is complete.
        """)
    datafile_cmd_update_parser = \
        datafile_command_parsers.add_parser("update",
                                            help=datafile_update_help,
                                            usage=datafile_update_usage)
    datafile_cmd_update_parser.add_argument(
        "datafile_id", help="The ID of the datafile to update.")
    datafile_cmd_update_parser.add_argument(
        "--md5sum", help="The new MD5 sum of the datafile.")



"""
Model class for the configuration,
usually stored in ~/.config/mytardisclient/mytardisclient.cfg
"""

# pylint: disable=missing-docstring

import os
import json
import traceback
from urlparse import urlparse
from ConfigParser import ConfigParser
from dogpile.cache import make_region  # pylint: disable=import-error

from mytardisclient.logs import logger

DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.config',
                                   'mytardisclient', 'mytardisclient.cfg')
DEFAULT_CACHE_PATH = os.path.join(os.path.expanduser('~'), '.cache',
                                  'mytardisclient', 'mytardisclient.cache')


class Config(object):
    """
    Model class for the minimal MyTardis server configuration
    (MyTardis URL, username and API key),
    usually stored in ~/.config/mytardisclient/mytardisclient.cfg
    """
    def __init__(self, path=DEFAULT_CONFIG_PATH):
        self.path = path
        self.url = ""
        self.username = ""
        self.apikey = ""
        self.default_headers = None
        def key_generator(namespace, function):
            # pylint: disable=unused-argument
            def generate_key(*args, **kwargs):
                return "%s(%s,%s)" % \
                    (function.__name__, str(args), str(kwargs))
            return generate_key
        self.cache_path = DEFAULT_CACHE_PATH
        if not os.path.exists(os.path.dirname(self.cache_path)):
            os.makedirs(os.path.dirname(self.cache_path))
        self.region = \
            make_region(function_key_generator=key_generator) \
                .configure(
                    'dogpile.cache.dbm',
                    expiration_time=30,
                    arguments={
                        "filename": self.cache_path
                    })
        if path:
            self.load()

    def __unicode__(self):
        return json.dumps(self.__dict__, indent=2)

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return self.__unicode__()

    def load(self, path=None):
        """
        Sets some default values for settings fields, then loads a config
        file, usually ~/.config/mytardisclient/mytardisclient.cfg
        """
        self.url = ""
        self.username = ""
        self.apikey = ""

        if path:
            self.path = path
        else:
            path = self.path

        if path is not None and os.path.exists(path):
            logger.info("Reading settings from: " + path)
            # pylint: disable=bare-except
            try:
                config_parser = ConfigParser()
                config_parser.read(path)
                section = "mytardisclient"
                fields = ["url", "username", "apikey"]

                # For backwards compatibility:
                if config_parser.has_option(section, "mytardis_url"):
                    self.url = config_parser.get(section, "mytardis_url")
                if config_parser.has_option(section, "api_key"):
                    self.apikey = config_parser.get(section, "api_key")

                for field in fields:
                    if config_parser.has_option(section, field):
                        self.__dict__[field] = \
                            config_parser.get(section, field)
            except:
                logger.error(traceback.format_exc())

        self.update_default_headers()

    def update_default_headers(self):
        self.default_headers = {
            "Authorization": "ApiKey %s:%s" % (self.username,
                                               self.apikey),
            "Content-Type": "application/json",
            "Accept": "application/json"
        }

    def validate(self):
        if self.username == "":
            raise Exception("MyTardis username is missing from config.")
        if self.apikey == "":
            raise Exception("MyTardis API key is missing from config.")
        if self.url == "":
            raise Exception("MyTardis URL is missing from config.")
        parsed_url = urlparse(self.url)
        if parsed_url.scheme not in ('http', 'https') or parsed_url.netloc == '':
            raise Exception("Invalid MyTardis URL found in config: %s", self.url)

    def save(self, path=None):
        if path:
            self.path = path
        else:
            path = self.path
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        config_parser = ConfigParser()
        with open(self.path, 'w') as config_file:
            config_parser.add_section("mytardisclient")
            fields = ["url", "username", "apikey"]
            for field in fields:
                config_parser.set("mytardisclient", field, self.__dict__[field])
            config_parser.write(config_file)
        logger.info("Saved settings to " + self.path)



"""
argparser/instrument.py
"""
import textwrap


def build_instrument_parser(argument_parser):
    """
    Builds parsing rules for instrument-related command-line interface arguments.
    """
    # pylint: disable=too-many-locals
    instrument_help = \
        "Display a list of instrument records or a single instrument record."
    instrument_usage = "mytardis instrument [-h] {list,get,create,update} ..."
    instrument_parser = \
        argument_parser.model_parsers.add_parser("instrument",
                                                 help=instrument_help,
                                                 usage=instrument_usage)
    instrument_command_parsers = \
        instrument_parser.add_subparsers(help='available commands',
                                         dest='command')

    instrument_list_help = "Display a list of instrument records."
    instrument_list_usage = textwrap.dedent("""\
        mytardis instrument list
            [--facility FACILITY] [--limit LIMIT] [--offset OFFSET] [--order_by ORDER_BY] [--json]

          EXAMPLE
       
          $ mytardis instrument list --facility 1

          Model: Instrument
          Query: http://mytardisdemo.erc.monash.edu.au/api/v1/instrument/?format=json&facility__id=1
          Total Count: 3
          Limit: 20
          Offset: 0

          +----+-------------------------+---------------+
          | ID |          Name           |   Facility    |
          +====+=========================+===============+
          |  3 | Test Instrument         | Demo Facility |
          +----+-------------------------+---------------+
          |  4 | Beamline                | Demo Facility |
          +----+-------------------------+---------------+
          |  8 | James Test Instrument   | Demo Facility |
          +----+-------------------------+---------------+
        """)
    instrument_command_list_parser = \
        instrument_command_parsers.add_parser("list",
                                              help=instrument_list_help,
                                              usage=instrument_list_usage)
    instrument_command_list_parser.add_argument("--facility",
                                                help="The facility ID.")
    instrument_command_list_parser.add_argument(
        "--limit", help="Maximum number of results to return.")
    instrument_command_list_parser.add_argument(
        "--offset",
        help="Skip this many records from the start of the result set.")
    instrument_command_list_parser.add_argument(
        "--order_by",
        help="Order by this field.")
    instrument_command_list_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    instrument_get_help = "Display a single instrument record."
    instrument_get_usage = textwrap.dedent("""\
        mytardis instrument get [-h] [--json] instrument_id

          EXAMPLE

          $ mytardis instrument get 3

          +------------------+-----------------+
          | Instrument field |      Value      |
          +==================+=================+
          | ID               | 3               |
          +------------------+-----------------+
          | Name             | Test Instrument |
          +------------------+-----------------+
          | Facility         | Demo Facility   |
          +------------------+-----------------+
        """)
    instrument_command_get_parser = \
        instrument_command_parsers.add_parser("get",
                                              help=instrument_get_help,
                                              usage=instrument_get_usage)
    instrument_command_get_parser.add_argument("instrument_id",
                                               help="The instrument ID.")
    instrument_command_get_parser.add_argument(
        "--json", action='store_true', help="Display results in JSON format.")

    instrument_create_help = "Create an instrument record."
    instrument_create_usage = textwrap.dedent("""\
        mytardis instrument create [-h] facility_id name

          EXAMPLE

          $ mytardis instrument create 1 "New Instrument"
          +------------------+----------------+
          | Instrument field |     Value      |
          +==================+================+
          | ID               | 9              |
          +------------------+----------------+
          | Name             | New Instrument |
          +------------------+----------------+
          | Facility         | Demo Facility  |
          +------------------+----------------+

          Instrument created successfully.
        """)
    instrument_cmd_create_parser = \
        instrument_command_parsers.add_parser("create",
                                              help=instrument_create_help,
                                              usage=instrument_create_usage)
    instrument_cmd_create_parser.add_argument(
        "facility_id", help="The ID of the new instrument's facility.")
    instrument_cmd_create_parser.add_argument(
        "name", help="The name of the instrument to create.")

    instrument_update_help = "Update/rename an existing instrument record."
    instrument_update_usage = textwrap.dedent("""\
        mytardis instrument update [-h] [--name NAME] instrument_id

          EXAMPLE

          $ mytardis instrument update --name "Renamed New Instrument" 9

          +------------------+------------------------+
          | Instrument field |         Value          |
          +==================+========================+
          | ID               | 9                      |
          +------------------+------------------------+
          | Name             | Renamed New Instrument |
          +------------------+------------------------+
          | Facility         | Demo Facility          |
          +------------------+------------------------+

          Instrument updated successfully.
        """)
    instrument_cmd_update_parser = \
        instrument_command_parsers.add_parser("update",
                                              help=instrument_update_help,
                                              usage=instrument_update_usage)
    instrument_cmd_update_parser.add_argument(
        "instrument_id", help="The ID of the instrument to update.")
    instrument_cmd_update_parser.add_argument(
        "--name", help="The new name of the instrument.")



"""
Controller class for running commands (list, get, create, update)
on instrument records.
"""

# from mytardisclient.logs import logger
from mytardisclient.models.instrument import Instrument
from mytardisclient.views import render


class InstrumentController(object):
    """
    Controller class for running commands (list, get, create, update)
    on instrument records.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.facility, args.limit,
                             args.offset, args.order_by,
                             render_format)
        elif command == "get":
            return self.get(args.instrument_id, render_format)
        elif command == "create":
            return self.create(args.facility_id, args.name, render_format)
        elif command == "update":
            return self.update(args.instrument_id, args.name, render_format)

    def list(self, facility_id, limit, offset, order_by, render_format):
        """
        Display list of instrument records.
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=no-self-use
        instruments = Instrument.list(facility_id,
                                      limit, offset, order_by)
        print render(instruments, render_format)

    def get(self, instrument_id, render_format):
        """
        Display instrument record.
        """
        # pylint: disable=no-self-use
        instrument = Instrument.get(instrument_id)
        print render(instrument, render_format)

    def create(self, facility_id, name, render_format):
        """
        Create instrument record.
        """
        # pylint: disable=no-self-use
        instrument = Instrument.create(facility_id, name)
        print render(instrument, render_format)
        print "Instrument created successfully."

    def update(self, instrument_id, name, render_format):
        """
        Update instrument record.
        """
        # pylint: disable=no-self-use
        instrument = Instrument.update(instrument_id, name)
        print render(instrument, render_format)
        print "Instrument updated successfully."



"""
Model class for MyTardis API v1's DatasetResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests
import json
import os

from mytardisclient.conf import config
from .resultset import ResultSet
from .schema import Schema
from .schema import ParameterName
from .instrument import Instrument
from mytardisclient.utils.exceptions import DoesNotExist


class Dataset(object):
    """
    Model class for MyTardis API v1's DatasetResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    def __init__(self, dataset_json=None):
        self.json = dataset_json
        self.id = None  # pylint: disable=invalid-name
        self.description = None
        self.instrument = None
        self.experiments = []
        self.parameter_sets = []
        if dataset_json:
            for key in self.__dict__.keys():
                if key in dataset_json:
                    self.__dict__[key] = dataset_json[key]
            if dataset_json['instrument']:
                self.instrument = Instrument(dataset_json['instrument'])
            self.parameter_sets = []
            for dataset_param_set_json in dataset_json['parameter_sets']:
                self.parameter_sets.append(DatasetParameterSet(dataset_param_set_json))

    @staticmethod
    @config.region.cache_on_arguments(namespace="Dataset")
    def list(experiment_id=None,
             limit=None, offset=None, order_by=None):
        """
        Get datasets I have access to
        """
        url = "%s/api/v1/dataset/?format=json" % config.url
        if experiment_id:
            url += "&experiments__id=%s"  % experiment_id
        if limit:
            url += "&limit=%s"  % limit
        if offset:
            url += "&offset=%s"  % offset
        if order_by:
            url += "&order_by=%s"  % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        if experiment_id or limit or offset:
            filters = dict(experiment_id=experiment_id,
                           limit=limit, offset=offset)
            return ResultSet(Dataset, url, response.json(), **filters)
        else:
            return ResultSet(Dataset, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="Dataset")
    def get(dataset_id):
        """
        Get dataset with id dataset_id
        """
        url = config.url + "/api/v1/dataset/?format=json" + "&id=%s" % dataset_id
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        datasets_json = response.json()
        if datasets_json['meta']['total_count'] == 0:
            message = "Dataset matching filter doesn't exist."
            raise DoesNotExist(message, url, response, Dataset)
        return Dataset(dataset_json=datasets_json['objects'][0])

    @staticmethod
    def create(experiment_id, description, instrument_id=None,
               params_file_json=None):
        """
        Create a dataset.
        """
        new_dataset_json = {
            "description": description,
            "experiments": ["/api/v1/experiment/%s/" % experiment_id],
            "immutable": False
        }
        if instrument_id:
            new_dataset_json['instrument'] = "/api/v1/instrument/%s/" \
                % instrument_id
        if params_file_json:
            assert os.path.exists(params_file_json)
            with open(params_file_json) as params_file:
                new_dataset_json['parameter_sets'] = json.load(params_file)
        url = config.url + "/api/v1/dataset/"
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_dataset_json))
        if response.status_code != 201:
            message = response.text
            raise Exception(message)
        dataset_json = response.json()
        return Dataset(dataset_json)

    @staticmethod
    def update(dataset_id, description):
        """
        Update an dataset record.
        """
        updated_fields_json = {'description': description}
        url = "%s/api/v1/dataset/%s/" % (config.url, dataset_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        if response.status_code != 202:
            print "HTTP %s" % response.status_code
            message = response.text
            raise Exception(message)
        dataset_json = response.json()
        return Dataset(dataset_json)


class DatasetParameterSet(object):
    """
    Model class for MyTardis API v1's DatasetParameterSetResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, dataset_paramset_json):
        self.json = dataset_paramset_json
        self.id = dataset_paramset_json['id']  # pylint: disable=invalid-name
        self.dataset = dataset_paramset_json['dataset']
        self.schema = Schema(dataset_paramset_json['schema'])
        self.parameters = []
        for dataset_param_json in dataset_paramset_json['parameters']:
            self.parameters.append(DatasetParameter(dataset_param_json))

    @staticmethod
    @config.region.cache_on_arguments(namespace="DatasetParameterSet")
    def list(dataset_id):
        """
        List dataset parameter sets associated with dataset ID
        dataset_id.
        """
        url = "%s/api/v1/datasetparameterset/?format=json" % config.url
        url += "&datasets__id=%s"  % dataset_id
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        filters = dict(dataset_id=dataset_id)
        return ResultSet(DatasetParameterSet, url, response.json(), **filters)


class DatasetParameter(object):
    """
    Model class for MyTardis API v1's DatasetParameterResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, dataset_param_json):
        self.json = dataset_param_json
        self.id = dataset_param_json['id']  # pylint: disable=invalid-name
        self.name = ParameterName.get(dataset_param_json['name'].split('/')[-2])
        self.string_value = dataset_param_json['string_value']
        self.numerical_value = dataset_param_json['numerical_value']
        self.datetime_value = dataset_param_json['datetime_value']
        self.link_id = dataset_param_json['link_id']
        self.value = dataset_param_json['value']

    @staticmethod
    @config.region.cache_on_arguments(namespace="DatasetParameter")
    def list(dataset_param_set):
        """
        List dataset parameter records in parameter set.
        """
        pass



"""
Controller class for running commands (list)
on storage boxes.
"""

from mytardisclient.models.storagebox import StorageBox
from mytardisclient.views import render


class StorageBoxController(object):
    """
    Controller class for running commands (list)
    on storage boxes.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.limit, args.offset, args.order_by,
                             render_format)
        elif command == "get":
            return self.get(args.storage_box_id, render_format)

    def list(self, limit, offset, order_by, render_format):
        """
        Display list of storage boxes.
        """
        # pylint: disable=no-self-use
        storage_boxes = StorageBox.list(limit, offset, order_by)
        print render(storage_boxes, render_format)

    def get(self, storage_box_id, render_format):
        """
        Display storage box record.
        """
        # pylint: disable=no-self-use
        storage_box = StorageBox.get(storage_box_id)
        print render(storage_box, render_format)



from setuptools import setup

import mytardisclient

setup(name='mytardisclient',
      packages=['mytardisclient'],
      version=mytardisclient.__version__,
      description='Command-line client for MyTardis API',
      author='James Wettenhall',
      author_email='james.wettenhall@monash.edu',
      url='http://github.com/wettenhj/mytardisclient',
      download_url = 'https://github.com/wettenhj/mytardisclient/archive/0.0.5.tar.gz',
      keywords = ['mytardis', 'REST'], # arbitrary keywords
      classifiers = [],
      license='GPL',
      entry_points={
          "console_scripts": [
              "mytardis = mytardisclient.client:run",
          ],
      },
      install_requires=['requests', 'ConfigParser', 'texttable',
                        'dogpile.cache'],
      zip_safe=False)



"""
Controller class for running commands (list, get, create, update)
on facility records.
"""

from mytardisclient.models.facility import Facility
from mytardisclient.models.instrument import Instrument
from mytardisclient.views import render


class FacilityController(object):
    """
    Controller class for running commands (list, get, create, update)
    on facility records.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.limit, args.offset, args.order_by,
                             render_format)
        if command == "get":
            return self.get(args.facility_id, render_format)

    def list(self, limit, offset, order_by, render_format):
        """
        Display list of facility records.
        """
        # pylint: disable=no-self-use
        facilities = Facility.list(limit, offset, order_by)
        print render(facilities, render_format)

    def get(self, facility_id, render_format):
        """
        Display facility record.
        """
        # pylint: disable=no-self-use
        facility = Facility.get(facility_id)
        print render(facility, render_format)
        if render_format == 'table':
            instruments = Instrument.list(facility_id)
            print render(instruments, render_format)



"""
Controller class for running commands (list, get, download, upload, update)
on datafile records.
"""

from mytardisclient.models.datafile import DataFile
from mytardisclient.views import render


class DataFileController(object):
    """
    Controller class for running commands (list, get, download, upload, update)
    on datafile records.
    """
    def __init__(self):
        pass

    def run_command(self, args):
        """
        Generic run command method.
        """
        command = args.command
        if hasattr(args, 'json') and args.json:
            render_format = 'json'
        else:
            render_format = 'table'
        if command == "list":
            return self.list(args.dataset, args.directory, args.filename,
                             args.limit, args.offset, args.order_by,
                             render_format)
        elif command == "get":
            return self.get(args.datafile_id, render_format)
        elif command == "create":
            return self.create(args.dataset_id, args.directory,
                               args.storagebox, args.file_path,
                               render_format)
        elif command == "download":
            return self.download(args.datafile_id)
        elif command == "upload":
            return self.upload(args.dataset_id, args.directory, args.file_path)
        elif command == "update":
            return self.update(args.datafile_id, args.md5sum, render_format)

    def list(self, dataset_id, directory, filename,
             limit, offset, order_by, render_format):
        """
        Display list of datafile records.
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=no-self-use
        datafiles = DataFile.list(dataset_id, directory, filename,
                                  limit, offset, order_by)
        print render(datafiles, render_format)

    def get(self, datafile_id, render_format):
        """
        Display datafile record.
        """
        # pylint: disable=no-self-use
        datafile = DataFile.get(datafile_id)
        print render(datafile, render_format)

    def create(self, dataset_id, directory, storagebox, file_path,
               render_format):
        """
        Create datafile record for an existing datafile.
        """
        # pylint: disable=too-many-arguments
        # pylint: disable=no-self-use
        datafile = DataFile.create(dataset_id, directory,
                                   storagebox, file_path)
        print render(datafile, render_format)
        print "DataFile created successfully."

    def download(self, datafile_id):
        """
        Download datafile.
        """
        # pylint: disable=no-self-use
        DataFile.download(datafile_id)

    def upload(self, dataset_id, directory, file_path):
        """
        Upload datafile.
        """
        # pylint: disable=no-self-use
        DataFile.upload(dataset_id, directory, file_path)

    def update(self, datafile_id, md5sum, render_format):
        """
        Update datafile record.
        """
        # pylint: disable=no-self-use
        datafile = DataFile.update(datafile_id, md5sum)
        print render(datafile, render_format)
        print "DataFile updated successfully."



"""
Model class for MyTardis API v1's SchemaResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests

from mytardisclient.conf import config
from .resultset import ResultSet
from mytardisclient.utils.exceptions import DoesNotExist


class Schema(object):
    """
    Model class for MyTardis API v1's SchemaResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, schema_json):
        self.json = schema_json
        self.id = schema_json['id']  # pylint: disable=invalid-name
        self.name = schema_json['name']
        self.hidden = schema_json['hidden']
        self.immutable = schema_json['immutable']
        self.namespace = schema_json['namespace']
        type_index = schema_json['type']
        _schema_types = ['', 'Experiment schema', 'Dataset schema', 'Datafile schema',
                         'None', 'Instrument schema']
        self.type = _schema_types[type_index]  # pylint: disable=invalid-name
        self.subtype = schema_json['subtype']

        self.parameter_names = ParameterName.list(schema=self)

    def __str__(self):
        return self.name

    @staticmethod
    @config.region.cache_on_arguments(namespace="Schema")
    def list(limit=None, offset=None, order_by=None):
        """
        Get schemas I have access to
        """
        url = "%s/api/v1/schema/?format=json" % config.url
        if limit:
            url += "&limit=%s" % limit
        if offset:
            url += "&offset=%s" % offset
        if order_by:
            url += "&order_by=%s" % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "URL: %s" % url
            print "HTTP %s" % response.status_code
            message = response.text
            raise Exception(message)

        if limit or offset:
            filters = dict(limit=limit, offset=offset)
            return ResultSet(Schema, url, response.json(), **filters)
        else:
            return ResultSet(Schema, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="Schema")
    def get(schema_id):
        """
        Get schema with id schema_id
        """
        url = "%s/api/v1/schema/%s/?format=json" % (config.url, schema_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            if response.status_code == 404:
                message = "Schema ID %s doesn't exist." % schema_id
                raise DoesNotExist(message, url, response, Schema)
            message = response.text
            raise Exception(message)

        schema_json = response.json()
        return Schema(schema_json=schema_json)


class ParameterName(object):
    """
    Model class for MyTardis API v1's ParameterNameResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, parametername_json):
        self.json = parametername_json
        self.schema = Schema.get(parametername_json['schema'].split('/')[-2])
        self.id = parametername_json['id']  # pylint: disable=invalid-name
        self.name = parametername_json['name']
        self.full_name = parametername_json['full_name']
        _type_choices = ['', 'Numeric', 'String', 'URL', 'Link',
                         'Filename', 'DateTime', 'Long String', 'JSON']
        self.data_type = _type_choices[parametername_json['data_type']]
        self.units = parametername_json['units']
        self.immutable = parametername_json['immutable']
        self.is_searchable = parametername_json['is_searchable']
        self.order = parametername_json['order']
        self.choices = parametername_json['choices']
        _comparison_types = \
            ['', 'Exact value', 'Not equal',
             'Range', 'Greater than', 'Greater than or equal to',
             'Less than', 'Less than or equal to', 'Contains']
        self.comparison_type = \
            _comparison_types[parametername_json['comparison_type']]

    def __unicode__(self):
        return self.full_name

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return self.__unicode__()

    @staticmethod
    @config.region.cache_on_arguments(namespace="ParameterName")
    def list(schema):
        """
        List parameter name records in schema.
        """
        # We would use "&schema__id=" if it were supported by MyTardis:
        url = "%s/api/v1/parametername/?format=json" % config.url
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            print "URL: %s" % url
            print "HTTP %s" % response.status_code
            message = response.text
            raise Exception(message)
        parameter_names_json = response.json()
        num_records = len(parameter_names_json['objects'])

        schema_resource_uri = "/api/v1/schema/%s/" % schema.id
        parameter_names_json['objects'] = \
            [pn for pn in parameter_names_json['objects']
             if pn['schema'] == schema_resource_uri]

        offset = 0
        limit = parameter_names_json['meta']['limit']
        total_count = parameter_names_json['meta']['total_count']
        while num_records < total_count:
            offset += limit
            url = "%s/api/v1/parametername/?format=json" % config.url
            url += "&offset=%s" % offset
            response = requests.get(url=url, headers=config.default_headers)
            if response.status_code != 200:
                print "URL: %s" % url
                print "HTTP %s" % response.status_code
                message = response.text
                raise Exception(message)
            parameter_names_page_json = response.json()
            num_records += len(parameter_names_page_json['objects'])
            parameter_names_page_json['objects'] = \
                [pn for pn in parameter_names_page_json['objects']
                 if pn['schema'] == schema_resource_uri]
            parameter_names_json['objects'].extend(parameter_names_page_json['objects'])

        return ResultSet(ParameterName, url, parameter_names_json)

    @staticmethod
    @config.region.cache_on_arguments(namespace="ParameterName")
    def get(parametername_id):
        """
        Get parameter name with id parametername_id
        """
        url = "%s/api/v1/parametername/%s/?format=json" % (config.url,
                                                           parametername_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            if response.status_code == 404:
                message = "Parameter Name ID %s doesn't exist." % parametername_id
                raise DoesNotExist(message, url, response, ParameterName)
            message = response.text
            raise Exception(message)

        parametername_json = response.json()
        return ParameterName(parametername_json=parametername_json)



"""
argparser/__init__.py
"""
from argparse import ArgumentParser
import textwrap

from .api import build_api_parser
from .config import build_config_parser
from .version import build_version_parser
from .facility import build_facility_parser
from .instrument import build_instrument_parser
from .experiment import build_experiment_parser
from .dataset import build_dataset_parser
from .datafile import build_datafile_parser
from .storagebox import build_storagebox_parser
from .schema import build_schema_parser


class ArgParser(object):
    """
    Defines parsing rules for command-line interface arguments.
    """
    def __init__(self):
        description = "Command-line interface for MyTardis REST API."
        self.parser = ArgumentParser(prog='mytardis', description=description)
        self.parser.add_argument(
            "--verbose", action='store_true', help="More verbose output.")
        self.model_parsers = \
            self.parser.add_subparsers(help='available models', dest='model')

    def get_args(self):
        """
        Builds argument parser and retrieves arguments.
        """
        self.build_parser()
        args = self.parser.parse_args()

        if args.model not in ('api', 'config', 'version',
                              'facility', 'instrument',
                              'experiment', 'dataset', 'datafile',
                              'storagebox', 'schema'):
            self.parser.error(
                "model should be one of 'api', 'config', 'version', "
                "'facility', 'instrument', "
                "'experiment', 'dataset', 'datafile', 'storagebox', "
                "'schema'.")

        return args

    def build_parser(self):
        """
        Builds parsing rules for command-line interface arguments.
        """
        build_api_parser(self)
        build_config_parser(self)
        build_version_parser(self)
        build_facility_parser(self)
        build_instrument_parser(self)
        build_experiment_parser(self)
        build_dataset_parser(self)
        build_datafile_parser(self)
        build_storagebox_parser(self)
        build_schema_parser(self)

        return self.parser


def get_parser():
    """
    Used by sphinx-argparse.
    """
    return ArgParser().build_parser()



"""
Model class for MyTardis API v1's InstrumentResource.
See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
"""

import requests
import json

from mytardisclient.conf import config
from .facility import Facility
from .resultset import ResultSet
from mytardisclient.utils.exceptions import DoesNotExist


class Instrument(object):
    """
    Model class for MyTardis API v1's InstrumentResource.
    See: https://github.com/mytardis/mytardis/blob/3.7/tardis/tardis_portal/api.py
    """
    def __init__(self, instrument_json):
        self.id = instrument_json['id']  # pylint: disable=invalid-name
        self.name = instrument_json['name']
        self.json = instrument_json
        self.facility = Facility(instrument_json['facility'])

    def __str__(self):
        return self.name

    @staticmethod
    @config.region.cache_on_arguments(namespace="Instrument")
    def list(facility_id=None, limit=None, offset=None, order_by=None):
        """
        Get instruments in facility with ID facility_id.
        """
        url = "%s/api/v1/instrument/?format=json" % config.url
        if facility_id:
            url += "&facility__id=%s" % facility_id
        if limit:
            url += "&limit=%s" % limit
        if offset:
            url += "&offset=%s" % offset
        if order_by:
            url += "&order_by=%s" % order_by
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        if facility_id or limit or offset:
            filters = dict(facility_id=facility_id, limit=limit, offset=offset)
            return ResultSet(Instrument, url, response.json(),
                             **filters)
        else:
            return ResultSet(Instrument, url, response.json())

    @staticmethod
    @config.region.cache_on_arguments(namespace="Instrument")
    def get(instrument_id):
        """
        Get instrument with id instrument_id
        """
        url = "%s/api/v1/instrument/?format=json&id=%s" % \
            (config.url, instrument_id)
        response = requests.get(url=url, headers=config.default_headers)
        if response.status_code != 200:
            message = response.text
            raise Exception(message)

        instruments_json = response.json()
        if instruments_json['meta']['total_count'] == 0:
            message = "Instrument matching filter doesn't exist."
            raise DoesNotExist(message, url, response, Instrument)
        return Instrument(instrument_json=instruments_json['objects'][0])

    @staticmethod
    def create(facility_id, name):
        """
        Create an instrument.
        """
        new_instrument_json = {
            "name": name,
            "facility": "/api/v1/facility/%s/" % facility_id
        }
        url = config.url + "/api/v1/instrument/"
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_instrument_json))
        if response.status_code != 201:
            message = response.text
            raise Exception(message)
        instrument_json = response.json()
        return Instrument(instrument_json)

    @staticmethod
    def update(instrument_id, name):
        """
        Update an instrument record.
        """
        updated_fields_json = {
            "name": name,
        }
        url = "%s/api/v1/instrument/%s/" % (config.url, instrument_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        if response.status_code != 202:
            print "HTTP %s" % response.status_code
            message = response.text
            raise Exception(message)
        instrument_json = response.json()
        return Instrument(instrument_json)
